#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Juju and Pebble mocking

This module contains mocks for the Juju and Pebble APIs that are used by ops
to interact with the Juju controller and the Pebble service manager.
"""

import datetime
import io
import shutil
import uuid
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Set,
    TextIO,
    Tuple,
    Union,
    cast,
    get_args,
)

from ops import (
    JujuVersion,
    pebble,
    SecretInfo,
    SecretNotFoundError,
    RelationNotFoundError,
    SecretRotate,
    ModelError,
)
from ops._private.harness import ExecArgs, _TestingPebbleClient
from ops.jujucontext import _JujuContext
from ops.model import CloudSpec as CloudSpec_Ops
from ops.model import Port as Port_Ops
from ops.model import Secret as Secret_Ops  # lol
from ops.model import (
    _format_action_result_dict,
    _ModelBackend,
    _SettableStatusName,
)
from ops.pebble import Client, ExecError

from .errors import ActionMissingFromContextError
from .logger import logger as scenario_logger
from .state import (
    CharmType,
    CheckInfo,
    JujuLogLine,
    Mount,
    Network,
    PeerRelation,
    Relation,
    RelationBase,
    Storage,
    SubordinateRelation,
    _EntityStatus,
    _port_cls_by_protocol,
    _RawPortProtocolLiteral,
)

if TYPE_CHECKING:  # pragma: no cover
    from .context import Context
    from .state import Container as ContainerSpec
    from .state import Exec, Secret, State, _CharmSpec, _Event

logger = scenario_logger.getChild("mocking")


class _MockExecProcess:
    def __init__(
        self,
        change_id: int,
        args: ExecArgs,
        return_code: int,
        stdin: Optional[Union[TextIO, io.BytesIO]],
        stdout: Optional[Union[TextIO, io.BytesIO]],
        stderr: Optional[Union[TextIO, io.BytesIO]],
    ):
        self._change_id = change_id
        self._args = args
        self._return_code = return_code
        self._waited = False
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

    def __del__(self):
        if not self._waited:
            self._close_stdin()

    def _close_stdin(self):
        if self._args.stdin is None and self.stdin is not None:
            self.stdin.seek(0)
            self._args.stdin = self.stdin.read()

    def wait(self):
        """Wait for the (mock) process to finish."""
        self._close_stdin()
        self._waited = True
        if self._return_code != 0:
            raise ExecError(list(self._args.command), self._return_code, None, None)

    def wait_output(self):
        """Wait for the (mock) process to finish and return tuple of (stdout, stderr)."""
        self._close_stdin()
        self._waited = True
        stdout = self.stdout.read() if self.stdout is not None else None
        stderr = self.stderr.read() if self.stderr is not None else None
        if self._return_code != 0:
            raise ExecError(
                list(self._args.command),
                self._return_code,
                stdout,  # type: ignore
                stderr,  # type: ignore
            )
        return stdout, stderr

    def send_signal(self, sig: Union[int, str]) -> NoReturn:  # noqa: U100
        """Send the given signal to the (mock) process."""
        raise NotImplementedError()


_NOT_GIVEN = object()  # non-None default value sentinel


# pyright: reportIncompatibleMethodOverride=false
class _MockModelBackend(_ModelBackend):  # type: ignore
    def __init__(
        self,
        state: "State",
        event: "_Event",
        charm_spec: "_CharmSpec[CharmType]",
        context: "Context",
        juju_context: "_JujuContext",
    ):
        super().__init__(juju_context=juju_context)
        self._state = state
        self._event = event
        self._context = context
        self._charm_spec = charm_spec

    def opened_ports(self) -> Set[Port_Ops]:
        return {
            Port_Ops(protocol=port.protocol, port=port.port)
            for port in self._state.opened_ports
        }

    def open_port(
        self,
        protocol: "_RawPortProtocolLiteral",
        port: Optional[int] = None,
    ):
        port_ = _port_cls_by_protocol[protocol](port=port)  # type: ignore
        ports = set(self._state.opened_ports)
        if port_ not in ports:
            ports.add(port_)
        if ports != self._state.opened_ports:
            self._state._update_opened_ports(frozenset(ports))

    def close_port(
        self,
        protocol: "_RawPortProtocolLiteral",
        port: Optional[int] = None,
    ):
        _port = _port_cls_by_protocol[protocol](port=port)  # type: ignore
        ports = set(self._state.opened_ports)
        if _port in ports:
            ports.remove(_port)
        if ports != self._state.opened_ports:
            self._state._update_opened_ports(frozenset(ports))

    def get_pebble(self, socket_path: str) -> "Client":
        container_name = socket_path.split("/")[
            3
        ]  # /charm/containers/<container_name>/pebble.socket
        container_root = self._context._get_container_root(container_name)
        try:
            mounts = self._state.get_container(container_name).mounts
        except KeyError:
            # container not defined in state.
            mounts = {}

        return cast(
            Client,
            _MockPebbleClient(
                socket_path=socket_path,
                container_root=container_root,
                mounts=mounts,
                state=self._state,
                event=self._event,
                charm_spec=self._charm_spec,
                context=self._context,
                container_name=container_name,
            ),
        )

    def _get_relation_by_id(self, rel_id: int) -> "RelationBase":
        try:
            return self._state.get_relation(rel_id)
        except KeyError:
            raise RelationNotFoundError() from None

    def _get_secret(self, id: Optional[str] = None, label: Optional[str] = None):
        if self._context.juju_version < "3.0.2":
            raise ModelError(
                "secrets are only available in juju >= 3.0.2."
                "Set ``Context.juju_version`` to 3.0.2+ to use them.",
            )

        canonicalize_id = Secret_Ops._canonicalize_id

        if id:
            # in scenario, you can create Secret(id="foo"),
            # but ops.Secret will prepend a "secret:" prefix to that ID.
            # we allow getting secret by either version.
            secrets = [
                s
                for s in self._state.secrets
                if canonicalize_id(s.id, model_uuid=self._state.model.uuid)
                == canonicalize_id(id, model_uuid=self._state.model.uuid)
            ]
            if not secrets:
                raise SecretNotFoundError(id)
            return secrets[0]

        if label:
            try:
                return self._state.get_secret(label=label)
            except KeyError:
                raise SecretNotFoundError(label) from None

        # if all goes well, this should never be reached. ops.model.Secret will check upon
        # instantiation that either an id or a label are set, and raise a TypeError if not.
        raise RuntimeError("need id or label.")

    def _check_app_data_access(self, is_app: bool):
        if not isinstance(is_app, bool):
            raise TypeError("is_app parameter to relation_get must be a boolean")

        if not is_app:
            return

        version = JujuVersion(self._context.juju_version)
        if not version.has_app_data():
            raise RuntimeError(
                f"setting application data is not supported on Juju version {version}",
            )

    def relation_get(self, relation_id: int, member_name: str, is_app: bool):
        self._check_app_data_access(is_app)
        relation = self._get_relation_by_id(relation_id)
        if is_app and member_name == self.app_name:
            return relation.local_app_data
        if is_app:
            if isinstance(relation, PeerRelation):
                return relation.local_app_data
            if isinstance(relation, (Relation, SubordinateRelation)):
                return relation.remote_app_data
            raise TypeError("relation_get: unknown relation type")
        if member_name == self.unit_name:
            return relation.local_unit_data

        unit_id = int(member_name.split("/")[-1])
        return relation._get_databag_for_remote(unit_id)  # noqa

    def is_leader(self):
        return self._state.leader

    def status_get(self, *, is_app: bool = False):
        status = self._state.app_status if is_app else self._state.unit_status
        return {"status": status.name, "message": status.message}

    def relation_ids(self, relation_name: str):
        return [
            rel.id for rel in self._state.relations if rel.endpoint == relation_name
        ]

    def relation_list(self, relation_id: int) -> Tuple[str, ...]:
        relation = self._get_relation_by_id(relation_id)

        if isinstance(relation, PeerRelation):
            return tuple(
                f"{self.app_name}/{unit_id}" for unit_id in relation.peers_data
            )
        remote_name = self.relation_remote_app_name(relation_id)
        return tuple(
            f"{remote_name}/{unit_id}" for unit_id in relation._remote_unit_ids
        )

    def config_get(self):
        state_config = self._state.config.copy()  # dedup or we'll mutate the state!

        # add defaults
        charm_config = self._charm_spec.config
        if not charm_config:
            return state_config

        for key, value in charm_config["options"].items():
            # if it has a default, and it's not overwritten from State, use it:
            if key not in state_config:
                default_value = value.get("default", _NOT_GIVEN)
                if default_value is not _NOT_GIVEN:  # accept False as default value
                    state_config[key] = default_value

        return state_config  # full config

    def network_get(self, binding_name: str, relation_id: Optional[int] = None):
        # validation:
        extra_bindings = self._charm_spec.meta.get("extra-bindings", ())
        all_endpoints = self._charm_spec.get_all_relations()
        non_sub_relations = {
            name for name, meta in all_endpoints if meta.get("scope") != "container"
        }

        # - is binding_name a valid binding name?
        if binding_name in extra_bindings:
            logger.warning("extra-bindings is a deprecated feature")  # fyi

            # - verify that if the binding is an extra binding, we're not ignoring a relation_id
            if relation_id is not None:
                # this should not happen
                logger.error(
                    "cannot pass relation_id to network_get if the binding name is "
                    "that of an extra-binding. Extra-bindings are not mapped to relation IDs.",
                )
        elif binding_name == "juju-info":
            # implicit relation that always exists
            pass
        # - verify that the binding is a relation endpoint name, but not a subordinate one
        elif binding_name not in non_sub_relations:
            logger.error(
                f"cannot get network binding for {binding_name}: is not a valid relation "
                f"endpoint name nor an extra-binding.",
            )
            raise RelationNotFoundError()

        # We look in State.networks for an override. If not given, we return a default network.
        try:
            network = self._state.get_network(binding_name)
        except KeyError:
            network = Network("default")  # The name is not used in the output.
        return network._hook_tool_output_fmt()

    # setter methods: these can mutate the state.
    def application_version_set(self, version: str):
        if workload_version := self._state.workload_version:
            # do not record if empty = unset
            self._context.workload_version_history.append(workload_version)

        self._state._update_workload_version(version)

    def status_set(
        self,
        status: _SettableStatusName,
        message: str = "",
        *,
        is_app: bool = False,
    ):
        valid_names = get_args(_SettableStatusName)
        if status not in valid_names:
            raise ModelError(
                f'ERROR invalid status "{status}", expected one of [{", ".join(valid_names)}]',
            )
        self._context._record_status(self._state, is_app)
        status_obj = _EntityStatus.from_status_name(status, message)
        self._state._update_status(status_obj, is_app)

    def juju_log(self, level: str, message: str):
        self._context.juju_log.append(JujuLogLine(level, message))

    def relation_set(self, relation_id: int, key: str, value: str, is_app: bool):
        self._check_app_data_access(is_app)
        relation = self._get_relation_by_id(relation_id)
        if is_app:
            if not self._state.leader:
                # will in practice not be reached because RelationData will check leadership
                # and raise RelationDataAccessError upstream on this path
                raise RuntimeError("needs leadership to set app data")
            tgt = relation.local_app_data
        else:
            tgt = relation.local_unit_data
        tgt[key] = value

    def secret_add(
        self,
        content: Dict[str, str],
        *,
        label: Optional[str] = None,
        description: Optional[str] = None,
        expire: Optional[datetime.datetime] = None,
        rotate: Optional[SecretRotate] = None,
        owner: Optional[Literal["unit", "app"]] = None,
    ) -> str:
        from .state import Secret

        secret = Secret(
            content,
            label=label,
            description=description,
            expire=expire,
            rotate=rotate,
            owner=owner,
        )
        secrets = set(self._state.secrets)
        secrets.add(secret)
        self._state._update_secrets(frozenset(secrets))
        return secret.id

    def _check_can_manage_secret(
        self,
        secret: "Secret",
    ):
        if secret.owner is None:
            raise SecretNotFoundError(
                "this secret is not owned by this unit/app or granted to it. "
                "Did you forget passing it to State.secrets?",
            )
        if secret.owner == "app" and not self.is_leader():
            understandable_error = SecretNotFoundError(
                f"App-owned secret {secret.id!r} can only be "
                f"managed by the leader.",
            )
            # charm-facing side: respect ops error
            raise ModelError("ERROR permission denied") from understandable_error

    def secret_get(
        self,
        *,
        id: Optional[str] = None,
        label: Optional[str] = None,
        refresh: bool = False,
        peek: bool = False,
    ) -> Dict[str, str]:
        secret = self._get_secret(id, label)
        # If both the id and label are provided, then update the label.
        if id is not None and label is not None:
            secret._set_label(label)
        juju_version = self._context.juju_version
        if not (juju_version == "3.1.7" or juju_version >= "3.3.1"):
            # In this medieval Juju chapter,
            # secret owners always used to track the latest revision.
            # ref: https://bugs.launchpad.net/juju/+bug/2037120
            if secret.owner is not None:
                refresh = True

        if peek or refresh:
            if refresh:
                secret._track_latest_revision()
            assert secret.latest_content is not None
            return secret.latest_content

        return secret.tracked_content

    def secret_info_get(
        self,
        *,
        id: Optional[str] = None,
        label: Optional[str] = None,
    ) -> SecretInfo:
        secret = self._get_secret(id, label)
        # If both the id and label are provided, then update the label.
        if id is not None and label is not None:
            secret._set_label(label)

        # only "manage"=write access level can read secret info
        self._check_can_manage_secret(secret)

        return SecretInfo(
            id=secret.id,
            label=secret.label,
            revision=secret._latest_revision,
            expires=secret.expire,
            rotation=secret.rotate,
            rotates=None,  # not implemented yet.
            model_uuid=self._state.model.uuid,
        )

    def secret_set(
        self,
        id: str,
        *,
        content: Optional[Dict[str, str]] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        expire: Optional[datetime.datetime] = None,
        rotate: Optional[SecretRotate] = None,
    ):
        secret = self._get_secret(id, label)
        self._check_can_manage_secret(secret)

        if content == secret.latest_content:
            # In Juju 3.6 and higher, this is a no-op, but it's good to warn
            # charmers if they are doing this, because it's not generally good
            # practice.
            # https://bugs.launchpad.net/juju/+bug/2069238
            logger.warning(
                f"secret {id} contents set to the existing value: new revision not needed",
            )

        secret._update_metadata(
            content=content,
            label=label,
            description=description,
            expire=expire,
            rotate=rotate,
        )

    def secret_grant(self, id: str, relation_id: int, *, unit: Optional[str] = None):
        secret = self._get_secret(id)
        self._check_can_manage_secret(secret)

        grantee = unit or self.relation_remote_app_name(
            relation_id,
            _raise_on_error=True,
        )

        if not secret.remote_grants.get(relation_id):
            secret.remote_grants[relation_id] = set()

        secret.remote_grants[relation_id].add(cast(str, grantee))

    def secret_revoke(self, id: str, relation_id: int, *, unit: Optional[str] = None):
        secret = self._get_secret(id)
        self._check_can_manage_secret(secret)

        grantee = unit or self.relation_remote_app_name(
            relation_id,
            _raise_on_error=True,
        )
        secret.remote_grants[relation_id].remove(cast(str, grantee))
        if not secret.remote_grants[relation_id]:
            del secret.remote_grants[relation_id]

    def secret_remove(self, id: str, *, revision: Optional[int] = None):
        secret = self._get_secret(id)
        self._check_can_manage_secret(secret)

        # Removing all revisions means that the secret is removed, so is no
        # longer in the state.
        if revision is None:
            secrets = set(self._state.secrets)
            secrets.remove(secret)
            self._state._update_secrets(frozenset(secrets))
            return

        # Juju does not prevent removing the tracked or latest revision, but it
        # is always a mistake to do this. Rather than having the state model a
        # secret where the tracked/latest revision cannot be retrieved but the
        # secret still exists, we raise instead, so that charms know that there
        # is a problem with their code.
        if revision in (secret._tracked_revision, secret._latest_revision):
            raise ValueError(
                "Charms should not remove the latest revision of a secret. "
                "Add a new revision with `set_content()` instead, and the previous "
                "revision will be cleaned up by the secret owner when no longer in use.",
            )

        # For all other revisions, the content is not visible to the charm
        # (this is as designed: the secret is being removed, so it should no
        # longer be in use). That means that the state does not need to be
        # modified - however, unit tests should be able to verify that the remove call was
        # executed, so we track that in a history list in the context.
        self._context.removed_secret_revisions.append(revision)

    def relation_remote_app_name(
        self,
        relation_id: int,
        _raise_on_error: bool = False,
    ) -> Optional[str]:
        # ops catches RelationNotFoundErrors and returns None:
        try:
            relation = self._get_relation_by_id(relation_id)
        except RelationNotFoundError:
            if _raise_on_error:
                raise
            return None

        if isinstance(relation, PeerRelation):
            return self.app_name
        if isinstance(relation, (Relation, SubordinateRelation)):
            return relation.remote_app_name
        raise TypeError("relation_remote_app_name: unknown relation type")

    def action_set(self, results: Dict[str, Any]):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-set",
            )
        # let ops validate the results dict
        _format_action_result_dict(results)
        # but then we will store it in its unformatted,
        # original form for testing ease
        if self._context.action_results:
            self._context.action_results.update(results)
        else:
            self._context.action_results = results

    def action_fail(self, message: str = ""):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-fail",
            )
        self._context._action_failure_message = message

    def action_log(self, message: str):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-log",
            )
        self._context.action_logs.append(message)

    def action_get(self):
        action = self._event.action
        if not action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-get",
            )
        return action.params

    def storage_add(self, name: str, count: int = 1):
        if not isinstance(count, int) or isinstance(count, bool):
            raise TypeError(
                f"storage count must be integer, got: {count} ({type(count)})",
            )

        if "/" in name:
            # this error is raised by Harness but not by ops at runtime
            raise ModelError('storage name cannot contain "/"')

        self._context.requested_storages[name] = count

    def storage_list(self, name: str) -> List[int]:
        return [
            storage.index for storage in self._state.storages if storage.name == name
        ]

    def _storage_event_details(self) -> Tuple[int, str]:
        storage = self._event.storage
        if not storage:
            # only occurs if this method is called when outside the scope of a storage event
            raise RuntimeError('unable to find storage key in ""')
        fs_path = storage.get_filesystem(self._context)
        return storage.index, str(fs_path)

    def storage_get(self, storage_name_id: str, attribute: str) -> str:
        if not len(attribute) > 0:  # assume it's an empty string.
            raise RuntimeError(
                'calling storage_get with `attribute=""` will return a dict '
                "and not a string. This usage is not supported.",
            )

        if attribute != "location":
            # this should not happen: in ops it's hardcoded to be "location"
            raise NotImplementedError(
                f"storage-get not implemented for attribute={attribute}",
            )

        name, index = storage_name_id.split("/")
        index = int(index)
        storages: List[Storage] = [
            s for s in self._state.storages if s.name == name and s.index == index
        ]

        # should not really happen: sanity checks. In practice, ops will guard against these paths.
        if not storages:
            raise RuntimeError(f"Storage with name={name} and index={index} not found.")
        if len(storages) > 1:
            raise RuntimeError(
                f"Multiple Storage instances with name={name} and index={index} found. "
                f"Inconsistent state.",
            )

        storage = storages[0]
        fs_path = storage.get_filesystem(self._context)
        return str(fs_path)

    def planned_units(self) -> int:
        return self._state.planned_units

    # legacy ops API that we don't intend to mock:
    def pod_spec_set(
        self,
        spec: Mapping[str, Any],  # noqa: U100
        k8s_resources: Optional[Mapping[str, Any]] = None,  # noqa: U100
    ) -> NoReturn:
        raise NotImplementedError(
            "pod-spec-set is not implemented in Scenario (and probably never will be: "
            "it's deprecated API)",
        )

    def add_metrics(
        self,
        metrics: Mapping[str, Union[int, float]],  # noqa: U100
        labels: Optional[Mapping[str, str]] = None,  # noqa: U100
    ) -> NoReturn:
        raise NotImplementedError(
            "add-metrics is not implemented in Scenario (and probably never will be: "
            "it's deprecated API)",
        )

    def resource_get(self, resource_name: str) -> str:
        # We assume that there are few enough resources that a linear search
        # will perform well enough.
        for resource in self._state.resources:
            if resource.name == resource_name:
                return str(resource.path)
        # ops will not let us get there if the resource name is unknown from metadata.
        # but if the user forgot to add it in State, then we remind you of that.
        raise RuntimeError(
            f"Inconsistent state: "
            f"resource {resource_name} not found in State. please pass it.",
        )

    def credential_get(self) -> CloudSpec_Ops:
        if not self._context.app_trusted:
            raise ModelError(
                "ERROR charm is not trusted, initialise Context with `app_trusted=True`",
            )
        if not self._state.model.cloud_spec:
            raise ModelError(
                "ERROR cloud spec is empty, initialise it with "
                "`State(model=Model(..., cloud_spec=ops.CloudSpec(...)))`",
            )
        return self._state.model.cloud_spec._to_ops()


class _MockPebbleClient(_TestingPebbleClient):
    def __init__(
        self,
        socket_path: str,
        container_root: Path,
        mounts: Dict[str, Mount],
        *,
        state: "State",
        event: "_Event",
        charm_spec: "_CharmSpec[CharmType]",
        context: "Context",
        container_name: str,
    ):
        self._state = state
        self.socket_path = socket_path
        self._event = event
        self._charm_spec = charm_spec
        self._context = context
        self._container_name = container_name

        # wipe just in case
        if container_root.exists():
            # Path.rmdir will fail if root is nonempty
            shutil.rmtree(container_root)

        # initialize simulated filesystem
        container_root.mkdir(parents=True)
        for _, mount in mounts.items():
            path = Path(mount.location).parts
            mounting_dir = container_root.joinpath(*path[1:])
            mounting_dir.parent.mkdir(parents=True, exist_ok=True)
            mounting_dir.symlink_to(mount.source)

        self._root = container_root

        self._notices: Dict[Tuple[str, str], pebble.Notice] = {}
        self._last_notice_id = 0
        self._changes: Dict[str, pebble.Change] = {}

        # load any existing notices and check information from the state
        self._notices: Dict[Tuple[str, str], pebble.Notice] = {}
        self._check_infos: Dict[str, pebble.CheckInfo] = {}
        try:
            container = state.get_container(self._container_name)
        except KeyError:
            # The container is in the metadata but not in the state - perhaps
            # this is an install event, at which point the container doesn't
            # exist yet. This means there will be no notices or check infos.
            pass
        else:
            for notice in container.notices:
                if hasattr(notice.type, "value"):
                    notice_type = cast(pebble.NoticeType, notice.type).value
                else:
                    notice_type = str(notice.type)
                self._notices[notice_type, notice.key] = notice._to_ops()
            now = datetime.datetime.now()
            for check in container.check_infos:
                self._check_infos[check.name] = check._to_ops()
                kind = (
                    pebble.ChangeKind.PERFORM_CHECK.value
                    if check.status == pebble.CheckStatus.UP
                    else pebble.ChangeKind.RECOVER_CHECK.value
                )
                change = pebble.Change(
                    pebble.ChangeID(str(uuid.uuid4())),
                    kind,
                    summary=check.name,
                    status=pebble.ChangeStatus.DOING.value,
                    tasks=[],
                    ready=False,
                    err=None,
                    spawn_time=now,
                    ready_time=now,
                )
                self._changes[check.change_id] = change

    def get_plan(self) -> pebble.Plan:
        return self._container.plan

    def _update_state_check_infos(self):
        """Copy any new or changed check infos into the state."""
        infos: set[CheckInfo] = set()
        for info in self._check_infos.values():
            if isinstance(info.level, str):
                level = pebble.CheckLevel(info.level)
            else:
                level = info.level
            if isinstance(info.status, str):
                status = pebble.CheckStatus(info.status)
            else:
                status = info.status
            check_info = CheckInfo(
                name=info.name,
                level=level,
                startup=info.startup,
                status=status,
                failures=info.failures,
                threshold=info.threshold,
                change_id=info.change_id,
            )
            infos.add(check_info)
        object.__setattr__(self._container, "check_infos", frozenset(infos))

    def replan_services(self, timeout: float = 30.0, delay: float = 0.1):
        super().replan_services(timeout=timeout, delay=delay)
        self._update_state_check_infos()

    def add_layer(
        self,
        label: str,
        layer: Union[str, "pebble.LayerDict", pebble.Layer],
        *,
        combine: bool = False,
    ):
        super().add_layer(label, layer, combine=combine)
        self._update_state_check_infos()

    def start_checks(self, names: List[str]) -> List[str]:
        started = super().start_checks(names)
        self._update_state_check_infos()
        return started

    def stop_checks(self, names: List[str]) -> List[str]:
        stopped = super().stop_checks(names)
        self._update_state_check_infos()
        return stopped

    @property
    def _container(self) -> "ContainerSpec":
        container_name = self.socket_path.split("/")[-2]
        try:
            return next(
                filter(lambda x: x.name == container_name, self._state.containers),
            )
        except StopIteration:
            raise RuntimeError(
                f"container with name={container_name!r} not found. "
                f"Did you forget a Container, or is the socket path "
                f"{self.socket_path!r} wrong?",
            )

    @property
    def _layers(self) -> Dict[str, pebble.Layer]:
        return self._container.layers

    @property
    def _service_status(self) -> Dict[str, pebble.ServiceStatus]:
        return self._container.service_statuses

    # Based on a method of the same name from Harness.
    def _find_exec_handler(self, command: List[str]) -> Optional["Exec"]:
        handlers = {exec.command_prefix: exec for exec in self._container.execs}
        # Start with the full command and, each loop iteration, drop the last
        # element, until it matches one of the command prefixes in the execs.
        # This includes matching against the empty list, which will match any
        # command, if there is not a more specific match.
        for prefix_len in reversed(range(len(command) + 1)):
            command_prefix = tuple(command[:prefix_len])
            if command_prefix in handlers:
                return handlers[command_prefix]
        # None of the command prefixes in the execs matched the command, no
        # matter how much of it was used, so we have failed to find a handler.
        return None

    def exec(
        self,
        command: List[str],
        *,
        environment: Optional[Dict[str, str]] = None,
        working_dir: Optional[str] = None,
        timeout: Optional[float] = None,
        user_id: Optional[int] = None,
        user: Optional[str] = None,
        group_id: Optional[int] = None,
        group: Optional[str] = None,
        stdin: Optional[Union[str, bytes, TextIO]] = None,
        stdout: Optional[TextIO] = None,
        stderr: Optional[TextIO] = None,
        encoding: Optional[str] = "utf-8",
        combine_stderr: bool = False,
        **kwargs: Any,
    ):
        handler = self._find_exec_handler(command)
        if not handler:
            raise ExecError(
                command,
                127,
                "",
                f"mock for cmd {command} not found. Please patch out whatever "
                f"leads to the call, or pass to the Container {self._container.name} "
                f"a scenario.Exec mock for the command your charm is attempting "
                f"to run, such as "
                f"'Container(..., execs={{scenario.Exec({list(command)}, ...)}})'",
            )

        if stdin is None:
            proc_stdin = self._transform_exec_handler_output("", encoding)
        else:
            proc_stdin = None
            stdin = stdin.read() if hasattr(stdin, "read") else stdin  # type: ignore
        if stdout is None:
            proc_stdout = self._transform_exec_handler_output(handler.stdout, encoding)
        else:
            proc_stdout = None
            stdout.write(handler.stdout)
        if stderr is None:
            proc_stderr = self._transform_exec_handler_output(handler.stderr, encoding)
        else:
            proc_stderr = None
            stderr.write(handler.stderr)

        args = ExecArgs(
            command=command,
            environment=environment or {},
            working_dir=working_dir,
            timeout=timeout,
            user_id=user_id,
            user=user,
            group_id=group_id,
            group=group,
            stdin=stdin,  # type:ignore  # If None, will be replaced by proc_stdin.read() later.
            encoding=encoding,
            combine_stderr=combine_stderr,
        )
        try:
            self._context.exec_history[self._container_name].append(args)
        except KeyError:
            self._context.exec_history[self._container_name] = [args]

        change_id = handler._run()
        return cast(
            pebble.ExecProcess[Any],
            _MockExecProcess(
                change_id=change_id,
                args=args,
                return_code=handler.return_code,
                stdin=proc_stdin,
                stdout=proc_stdout,
                stderr=proc_stderr,
            ),
        )

    def _check_connection(self):
        if not self._container.can_connect:
            msg = (
                f"Cannot connect to Pebble; did you forget to set "
                f"can_connect=True for container {self._container.name}?"
            )
            raise pebble.ConnectionError(msg)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark tests for ops-scenario.

Optimising performance is not a current goal with ops-scenario. However,
we do want to ensure that we do not unknowingly regress in performance.

This package contains a small set of tests that cover core functionality,
to be used for performance benchmarking.
"""



import pytest
from ops import RelationNotFoundError
from ops.charm import CharmBase
from ops.framework import Framework

from scenario import Context
from scenario.state import (
    Address,
    BindAddress,
    Network,
    Relation,
    State,
    SubordinateRelation,
)


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        _call = None
        called = False

        def __init__(self, framework: Framework):
            super().__init__(framework)

            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if MyCharm._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_ip_get(mycharm):
    ctx = Context(
        mycharm,
        meta={
            "name": "foo",
            "requires": {
                "metrics-endpoint": {"interface": "foo"},
                "deadnodead": {"interface": "bar"},
            },
            "extra-bindings": {"foo": {}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                Relation(
                    interface="foo",
                    remote_app_name="remote",
                    endpoint="metrics-endpoint",
                    id=1,
                ),
            ],
            networks={Network("foo", [BindAddress([Address("4.4.4.4")])])},
        ),
    ) as mgr:
        # we have a network for the relation
        rel = mgr.charm.model.get_relation("metrics-endpoint")
        assert str(mgr.charm.model.get_binding(rel).network.bind_address) == "192.0.2.0"

        # we have a network for a binding without relations on it
        assert (
            str(mgr.charm.model.get_binding("deadnodead").network.bind_address)
            == "192.0.2.0"
        )

        # and an extra binding
        assert str(mgr.charm.model.get_binding("foo").network.bind_address) == "4.4.4.4"


def test_no_sub_binding(mycharm):
    ctx = Context(
        mycharm,
        meta={
            "name": "foo",
            "requires": {"bar": {"interface": "foo", "scope": "container"}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                SubordinateRelation("bar"),
            ]
        ),
    ) as mgr:
        with pytest.raises(RelationNotFoundError):
            # sub relations have no network
            mgr.charm.model.get_binding("bar").network


def test_no_relation_error(mycharm):
    """Attempting to call get_binding on a non-existing relation -> RelationNotFoundError"""

    ctx = Context(
        mycharm,
        meta={
            "name": "foo",
            "requires": {"metrics-endpoint": {"interface": "foo"}},
            "extra-bindings": {"bar": {}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                Relation(
                    interface="foo",
                    remote_app_name="remote",
                    endpoint="metrics-endpoint",
                    id=1,
                ),
            ],
            networks={Network("bar")},
        ),
    ) as mgr:
        with pytest.raises(RelationNotFoundError):
            mgr.charm.model.get_binding("foo").network


def test_juju_info_network_default(mycharm):
    ctx = Context(
        mycharm,
        meta={"name": "foo"},
    )

    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        # we have a network for the relation
        assert (
            str(mgr.charm.model.get_binding("juju-info").network.bind_address)
            == "192.0.2.0"
        )


def test_explicit_juju_info_network_override(mycharm):
    ctx = Context(
        mycharm,
        meta={
            "name": "foo",
            # this charm for whatever reason explicitly defines a juju-info endpoint
            "requires": {"juju-info": {"interface": "juju-info"}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        assert mgr.charm.model.get_binding("juju-info").network.bind_address



import os
from tempfile import TemporaryDirectory

import pytest

import ops

from scenario import Context
from scenario.state import Relation, State, _CharmSpec, _Event
from scenario._runtime import Runtime, UncaughtCharmError


def charm_type():
    class _CharmEvents(ops.CharmEvents):
        pass

    class MyCharm(ops.CharmBase):
        on = _CharmEvents()  # type: ignore
        _event = None

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._catchall)

        def _catchall(self, e: ops.EventBase):
            if self._event:
                return
            MyCharm._event = e

    return MyCharm


def test_event_emission():
    with TemporaryDirectory():
        meta = {
            "name": "foo",
            "requires": {"ingress-per-unit": {"interface": "ingress_per_unit"}},
        }

        my_charm_type = charm_type()

        class MyEvt(ops.EventBase):
            pass

        my_charm_type.on.define_event("bar", MyEvt)

        runtime = Runtime(
            _CharmSpec(
                my_charm_type,
                meta=meta,
            ),
        )

        with runtime.exec(
            state=State(),
            event=_Event("bar"),
            context=Context(my_charm_type, meta=meta),
        ) as manager:
            manager.run()

        assert my_charm_type._event
        assert isinstance(my_charm_type._event, MyEvt)


@pytest.mark.parametrize("app_name", ("foo", "bar-baz", "QuX2"))
@pytest.mark.parametrize("unit_id", (1, 2, 42))
def test_unit_name(app_name, unit_id):
    meta = {
        "name": app_name,
    }

    my_charm_type = charm_type()

    runtime = Runtime(
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
        unit_id=unit_id,
        app_name=app_name,
    )

    with runtime.exec(
        state=State(),
        event=_Event("start"),
        context=Context(my_charm_type, meta=meta),
    ) as manager:
        assert manager.charm.unit.name == f"{app_name}/{unit_id}"


def test_env_clean_on_charm_error():
    meta = {"name": "frank", "requires": {"box": {"interface": "triangle"}}}

    my_charm_type = charm_type()

    runtime = Runtime(
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
    )

    remote_name = "ava"
    rel = Relation("box", remote_app_name=remote_name)
    with pytest.raises(UncaughtCharmError) as exc:
        with runtime.exec(
            state=State(relations={rel}),
            event=_Event("box_relation_changed", relation=rel),
            context=Context(my_charm_type, meta=meta),
        ) as manager:
            assert manager._juju_context.remote_app_name == remote_name
            assert "JUJU_REMOTE_APP" in os.environ
            _ = 1 / 0  # raise some error
    # Ensure that some other error didn't occur (like AssertionError!).
    assert "ZeroDivisionError" in str(exc.value)

    # Ensure that the Juju environment didn't leak into the outside one.
    assert os.getenv("JUJU_REMOTE_APP", None) is None


def test_juju_version_is_set_in_environ():
    version = "2.9"

    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, _: ops.StartEvent):
            assert ops.JujuVersion.from_environ() == version

    ctx = Context(MyCharm, meta={"name": "foo"}, juju_version=version)
    ctx.run(ctx.on.start(), State())



import copy
from dataclasses import asdict, replace
from typing import Optional, Type

import ops
import pytest
from ops.charm import CharmBase, CharmEvents, CollectStatusEvent
from ops.framework import EventBase, Framework
from ops.model import ActiveStatus, UnknownStatus, WaitingStatus

from scenario.state import (
    _DEFAULT_JUJU_DATABAG,
    Address,
    BindAddress,
    CheckInfo,
    Container,
    Model,
    Network,
    Relation,
    Resource,
    State,
)
from tests.helpers import jsonpatch_delta, sort_patch, trigger

CUSTOM_EVT_SUFFIXES = {
    "relation_created",
    "relation_joined",
    "relation_changed",
    "relation_departed",
    "relation_broken",
    "storage_attached",
    "storage_detaching",
    "action",
    "pebble_ready",
    "pebble_custom_notice",
}


@pytest.fixture(scope="function")
def mycharm():
    class MyCharmEvents(CharmEvents):
        @classmethod
        def define_event(cls, event_kind: str, event_type: "Type[EventBase]"):
            if getattr(cls, event_kind, None):
                delattr(cls, event_kind)
            return super().define_event(event_kind, event_type)

    class MyCharm(CharmBase):
        _call = None
        called = False
        on = MyCharmEvents()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if self._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


@pytest.fixture
def state():
    return State(config={"foo": "bar"}, leader=True)


def test_bare_event(state, mycharm):
    out = trigger(
        state,
        "start",
        mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}}},
    )
    out_purged = replace(out, stored_states=state.stored_states)
    assert jsonpatch_delta(state, out_purged) == []


def test_leader_get(state, mycharm):
    def pre_event(charm):
        assert charm.unit.is_leader()

    trigger(
        state,
        "start",
        mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}}},
        pre_event=pre_event,
    )


def test_status_setting(state, mycharm):
    def call(charm: CharmBase, e):
        if isinstance(e, CollectStatusEvent):
            return

        assert isinstance(charm.unit.status, UnknownStatus)
        charm.unit.status = ActiveStatus("foo test")
        charm.app.status = WaitingStatus("foo barz")

    mycharm._call = call
    out = trigger(
        state,
        "start",
        mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}}},
    )
    assert out.unit_status == ActiveStatus("foo test")
    assert out.app_status == WaitingStatus("foo barz")
    assert out.workload_version == ""

    # ignore stored state in the delta
    out_purged = replace(out, stored_states=state.stored_states)
    assert jsonpatch_delta(out_purged, state) == sort_patch([
        {"op": "replace", "path": "/app_status/message", "value": "foo barz"},
        {"op": "replace", "path": "/app_status/name", "value": "waiting"},
        {"op": "replace", "path": "/unit_status/message", "value": "foo test"},
        {"op": "replace", "path": "/unit_status/name", "value": "active"},
    ])


@pytest.mark.parametrize("connect", (True, False))
def test_container(connect, mycharm):
    def pre_event(charm: CharmBase):
        container = charm.unit.get_container("foo")
        assert container is not None
        assert container.name == "foo"
        assert container.can_connect() is connect

    trigger(
        State(containers={Container(name="foo", can_connect=connect)}),
        "start",
        mycharm,
        meta={
            "name": "foo",
            "containers": {"foo": {"resource": "bar"}},
        },
        pre_event=pre_event,
    )


def test_relation_get(mycharm):
    def pre_event(charm: CharmBase):
        rel = charm.model.get_relation("foo")
        assert rel is not None
        assert rel.data[charm.app]["a"] == "because"

        assert rel.data[rel.app]["a"] == "b"
        assert rel.data[charm.unit]["c"] == "d"

        for unit in rel.units:
            if unit is charm.unit:
                continue
            if unit.name == "remote/1":
                assert rel.data[unit]["e"] == "f"
            else:
                assert not rel.data[unit]

    state = State(
        relations={
            Relation(
                endpoint="foo",
                interface="bar",
                local_app_data={"a": "because"},
                remote_app_name="remote",
                remote_app_data={"a": "b"},
                local_unit_data={"c": "d"},
                remote_units_data={0: {}, 1: {"e": "f"}, 2: {}},
            )
        }
    )
    trigger(
        state,
        "start",
        mycharm,
        meta={
            "name": "local",
            "requires": {"foo": {"interface": "bar"}},
        },
        pre_event=pre_event,
    )


def test_relation_set(mycharm):
    def event_handler(charm: CharmBase, _):
        rel = charm.model.get_relation("foo")
        rel.data[charm.app]["a"] = "b"
        rel.data[charm.unit]["c"] = "d"

        # this will NOT raise an exception because we're not in an event context!
        # we're right before the event context is entered in fact.
        with pytest.raises(Exception):
            rel.data[rel.app]["a"] = "b"
        with pytest.raises(Exception):
            rel.data[charm.model.get_unit("remote/1")]["c"] = "d"

        assert charm.unit.is_leader()

    def pre_event(charm: CharmBase):
        assert charm.model.get_relation("foo")
        assert charm.model.app.planned_units() == 4

        # this would NOT raise an exception because we're not in an event context!
        # we're right before the event context is entered in fact.
        # with pytest.raises(Exception):
        #     rel.data[rel.app]["a"] = "b"
        # with pytest.raises(Exception):
        #     rel.data[charm.model.get_unit("remote/1")]["c"] = "d"

    mycharm._call = event_handler
    relation = Relation(
        endpoint="foo",
        interface="bar",
        remote_app_name="remote",
        remote_units_data={1: {}, 4: {}},
    )
    state = State(
        leader=True,
        planned_units=4,
        relations={relation},
    )

    assert not mycharm.called
    out = trigger(
        state,
        event="start",
        charm_type=mycharm,
        meta={
            "name": "foo",
            "requires": {"foo": {"interface": "bar"}},
        },
        pre_event=pre_event,
    )
    assert mycharm.called

    assert asdict(out.get_relation(relation.id)) == asdict(
        replace(
            relation,
            local_app_data={"a": "b"},
            local_unit_data={"c": "d", **_DEFAULT_JUJU_DATABAG},
        )
    )
    assert out.get_relation(relation.id).local_app_data == {"a": "b"}
    assert out.get_relation(relation.id).local_unit_data == {
        "c": "d",
        **_DEFAULT_JUJU_DATABAG,
    }


def test_checkinfo_changeid_none():
    info = CheckInfo("foo", change_id=None)
    assert info.change_id, "None should result in a random change_id"
    info2 = CheckInfo("foo")  # None is also the default.
    assert info.change_id != info2.change_id


@pytest.mark.parametrize("id", ("", "28"))
def test_checkinfo_changeid(id: Optional[str]):
    info = CheckInfo("foo", change_id=ops.pebble.ChangeID(id))
    assert info.change_id == ops.pebble.ChangeID(id)


@pytest.mark.parametrize(
    "klass,num_args",
    [
        (State, (1,)),
        (Resource, (1,)),
        (Address, (0, 2)),
        (BindAddress, (0, 2)),
        (Network, (0, 3)),
    ],
)
def test_positional_arguments(klass, num_args):
    for num in num_args:
        args = (None,) * num
        with pytest.raises(TypeError):
            klass(*args)


def test_model_positional_arguments():
    with pytest.raises(TypeError):
        Model("", "")


def test_container_positional_arguments():
    with pytest.raises(TypeError):
        Container("", "")


def test_container_default_values():
    name = "foo"
    container = Container(name)
    assert container.name == name
    assert container.can_connect is False
    assert container.layers == {}
    assert container.service_statuses == {}
    assert container.mounts == {}
    assert container.execs == frozenset()
    assert container.layers == {}
    assert container._base_plan == {}


def test_state_default_values():
    state = State()
    assert state.config == {}
    assert state.relations == frozenset()
    assert state.networks == frozenset()
    assert state.containers == frozenset()
    assert state.storages == frozenset()
    assert state.opened_ports == frozenset()
    assert state.secrets == frozenset()
    assert state.resources == frozenset()
    assert state.deferred == []
    assert isinstance(state.model, Model)
    assert state.leader is False
    assert state.planned_units == 1
    assert state.app_status == UnknownStatus()
    assert state.unit_status == UnknownStatus()
    assert state.workload_version == ""


def test_deepcopy_state():
    containers = [Container("foo"), Container("bar")]
    state = State(containers=containers)
    state_copy = copy.deepcopy(state)
    for container in state.containers:
        copied_container = state_copy.get_container(container.name)
        assert container.name == copied_container.name


def test_replace_state():
    containers = [Container("foo"), Container("bar")]
    state = State(containers=containers, leader=True)
    state2 = replace(state, leader=False)
    assert state.leader != state2.leader
    assert state.containers == state2.containers



import sys

pytest_plugins = "pytester"
sys.path.append(".")


def test_plugin_ctx_run(pytester):
    # create a temporary pytest test module
    pytester.makepyfile(
        """
        import pytest
        from scenario import State
        from scenario import Context
        import ops

        class MyCharm(ops.CharmBase):
            pass

        @pytest.fixture
        def context():
            return Context(charm_type=MyCharm, meta={"name": "foo"})

        def test_sth(context):
            context.run(context.on.start(), State())
    """
    )

    # run pytest with the following cmd args
    result = pytester.runpytest("-v")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines([
        "*::test_sth PASSED*",
    ])

    # make sure that we get a '0' exit code for the testsuite
    assert result.ret == 0



import pytest
from ops import __version__ as ops_version
from ops.charm import ActionEvent, CharmBase
from ops.framework import Framework
from ops._private.harness import ActionFailed

from scenario import Context
from scenario.state import State, _Action, _next_action_id


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        _evt_handler = None

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if handler := self._evt_handler:
                handler(event)

    return MyCharm


@pytest.mark.parametrize("baz_value", (True, False))
def test_action_event(mycharm, baz_value):
    ctx = Context(
        mycharm,
        meta={"name": "foo"},
        actions={
            "foo": {"params": {"bar": {"type": "number"}, "baz": {"type": "boolean"}}}
        },
    )
    state = ctx.run(ctx.on.action("foo", params={"baz": baz_value, "bar": 10}), State())

    assert isinstance(state, State)
    evt = ctx.emitted_events[0]
    assert evt.params["bar"] == 10
    assert evt.params["baz"] is baz_value


def test_action_no_results():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.act_action, self._on_act_action)

        def _on_act_action(self, _):
            pass

    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"act": {}})
    ctx.run(ctx.on.action("act"), State())
    assert ctx.action_results is None
    assert ctx.action_logs == []


@pytest.mark.parametrize("res_value", ("one", 1, [2], ["bar"], (1,), {1, 2}))
def test_action_event_results_invalid(mycharm, res_value):
    def handle_evt(charm: CharmBase, evt: ActionEvent):
        with pytest.raises((TypeError, AttributeError)):
            evt.set_results(res_value)

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    ctx.run(ctx.on.action("foo"), State())


@pytest.mark.parametrize("res_value", ({"a": {"b": {"c"}}}, {"d": "e"}))
def test_action_event_results_valid(mycharm, res_value):
    def handle_evt(_: CharmBase, evt):
        if not isinstance(evt, ActionEvent):
            return
        evt.set_results(res_value)
        evt.log("foo")
        evt.log("bar")

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})

    ctx.run(ctx.on.action("foo"), State())

    assert ctx.action_results == res_value


@pytest.mark.parametrize("res_value", ({"a": {"b": {"c"}}}, {"d": "e"}))
def test_action_event_outputs(mycharm, res_value):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return

        evt.set_results({"my-res": res_value})
        evt.log("log1")
        evt.log("log2")
        evt.fail("failed becozz")

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    with pytest.raises(ActionFailed) as exc_info:
        ctx.run(ctx.on.action("foo"), State())
    assert exc_info.value.message == "failed becozz"
    assert ctx.action_results == {"my-res": res_value}
    assert ctx.action_logs == ["log1", "log2"]


def test_action_continues_after_fail():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_action, self._on_foo_action)

        def _on_foo_action(self, event):
            event.log("starting")
            event.set_results({"initial": "result"})
            event.fail("oh no!")
            event.set_results({"final": "result"})

    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}})
    with pytest.raises(ActionFailed) as exc_info:
        ctx.run(ctx.on.action("foo"), State())
    assert exc_info.value.message == "oh no!"
    assert ctx.action_logs == ["starting"]
    assert ctx.action_results == {"initial": "result", "final": "result"}


def _ops_less_than(wanted_major, wanted_minor):
    major, minor = (int(v) for v in ops_version.split(".")[:2])
    if major < wanted_major:
        return True
    if major == wanted_major and minor < wanted_minor:
        return True
    return False


@pytest.mark.skipif(
    _ops_less_than(2, 11), reason="ops 2.10 and earlier don't have ActionEvent.id"
)
def test_action_event_has_id(mycharm):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        assert isinstance(evt.id, str) and evt.id != ""

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    ctx.run(ctx.on.action("foo"), State())


@pytest.mark.skipif(
    _ops_less_than(2, 11), reason="ops 2.10 and earlier don't have ActionEvent.id"
)
def test_action_event_has_override_id(mycharm):
    uuid = "0ddba11-cafe-ba1d-5a1e-dec0debad"

    def handle_evt(charm: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        assert evt.id == uuid

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    ctx.run(ctx.on.action("foo", id=uuid), State())


def test_two_actions_same_context():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_action, self._on_foo_action)
            framework.observe(self.on.bar_action, self._on_bar_action)

        def _on_foo_action(self, event):
            event.log("foo")
            event.set_results({"foo": "result"})

        def _on_bar_action(self, event):
            event.log("bar")
            event.set_results({"bar": "result"})

    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}, "bar": {}})
    ctx.run(ctx.on.action("foo"), State())
    assert ctx.action_results == {"foo": "result"}
    assert ctx.action_logs == ["foo"]
    # Not recommended, but run another action in the same context.
    ctx.run(ctx.on.action("bar"), State())
    assert ctx.action_results == {"bar": "result"}
    assert ctx.action_logs == ["bar"]


def test_positional_arguments():
    with pytest.raises(TypeError):
        _Action("foo", {})


def test_default_arguments():
    expected_id = _next_action_id(update=False)
    name = "foo"
    action = _Action(name)
    assert action.name == name
    assert action.params == {}
    assert action.id == expected_id



import os

import pytest
from ops.charm import CharmBase, CharmEvents
from ops.framework import EventBase, EventSource, Framework, Object

from scenario.state import State, _CharmSpec, _Event
from ..helpers import trigger


class QuxEvent(EventBase):
    pass


class SubEvent(EventBase):
    pass


@pytest.fixture(scope="function")
def mycharm():
    class MyCharmEvents(CharmEvents):
        qux = EventSource(QuxEvent)

    class MySubEvents(CharmEvents):
        sub = EventSource(SubEvent)

    class Sub(Object):
        on = MySubEvents()

    class MyCharm(CharmBase):
        on = MyCharmEvents()
        evts = []

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.sub = Sub(self, "sub")
            self.framework.observe(self.sub.on.sub, self._on_event)
            self.framework.observe(self.on.qux, self._on_event)

        def _on_event(self, e):
            MyCharm.evts.append(e)

    return MyCharm


@pytest.mark.parametrize("evt_name", ("rubbish", "foo", "bar", "kazoo_pebble_ready"))
def test_rubbish_event_raises(mycharm, evt_name):
    with pytest.raises(AttributeError):
        if evt_name.startswith("kazoo"):
            os.environ["SCENARIO_SKIP_CONSISTENCY_CHECKS"] = "true"
            # else it will whine about the container not being in state and meta;
            # but if we put the container in meta, it will actually register an event!

        trigger(State(), evt_name, mycharm, meta={"name": "foo"})

        if evt_name.startswith("kazoo"):
            os.environ["SCENARIO_SKIP_CONSISTENCY_CHECKS"] = "false"


@pytest.mark.parametrize("evt_name", ("qux",))
def test_custom_events_fail(mycharm, evt_name):
    with pytest.raises(AttributeError):
        trigger(State(), evt_name, mycharm, meta={"name": "foo"})


# cfr: https://github.com/PietroPasotti/ops-scenario/pull/11#discussion_r1101694961
@pytest.mark.parametrize("evt_name", ("sub",))
def test_custom_events_sub_raise(mycharm, evt_name):
    with pytest.raises(AttributeError):
        trigger(State(), evt_name, mycharm, meta={"name": "foo"})


@pytest.mark.parametrize(
    "evt_name, expected",
    (
        ("qux", False),
        ("sub", False),
        ("start", True),
        ("install", True),
        ("config-changed", True),
        ("foo-relation-changed", True),
        ("bar-relation-changed", True),
    ),
)
def test_is_custom_event(mycharm, evt_name, expected):
    spec = _CharmSpec(
        charm_type=mycharm, meta={"name": "mycharm", "requires": {"foo": {}}}
    )
    assert _Event(evt_name)._is_builtin_event(spec) is expected



import ops
import pytest

import scenario


class MyCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        for evt in self.on.events().values():
            self.framework.observe(evt, self._on_event)

    def _on_event(self, event):
        pass


def test_get_cloud_spec():
    scenario_cloud_spec = scenario.CloudSpec(
        type="lxd",
        name="localhost",
        endpoint="https://127.0.0.1:8443",
        credential=scenario.CloudCredential(
            auth_type="clientcertificate",
            attributes={
                "client-cert": "foo",
                "client-key": "bar",
                "server-cert": "baz",
            },
        ),
    )
    expected_cloud_spec = ops.CloudSpec(
        type="lxd",
        name="localhost",
        endpoint="https://127.0.0.1:8443",
        credential=ops.CloudCredential(
            auth_type="clientcertificate",
            attributes={
                "client-cert": "foo",
                "client-key": "bar",
                "server-cert": "baz",
            },
        ),
    )
    ctx = scenario.Context(MyCharm, meta={"name": "foo"}, app_trusted=True)
    state = scenario.State(
        model=scenario.Model(
            name="lxd-model", type="lxd", cloud_spec=scenario_cloud_spec
        ),
    )
    with ctx(ctx.on.start(), state=state) as mgr:
        assert mgr.charm.model.get_cloud_spec() == expected_cloud_spec


def test_get_cloud_spec_error():
    ctx = scenario.Context(MyCharm, meta={"name": "foo"})
    state = scenario.State(model=scenario.Model(name="lxd-model", type="lxd"))
    with ctx(ctx.on.start(), state) as mgr:
        with pytest.raises(ops.ModelError):
            mgr.charm.model.get_cloud_spec()


def test_get_cloud_spec_untrusted():
    cloud_spec = ops.CloudSpec(type="lxd", name="localhost")
    ctx = scenario.Context(MyCharm, meta={"name": "foo"})
    state = scenario.State(
        model=scenario.Model(name="lxd-model", type="lxd", cloud_spec=cloud_spec),
    )
    with ctx(ctx.on.start(), state) as mgr:
        with pytest.raises(ops.ModelError):
            mgr.charm.model.get_cloud_spec()



"""pytest configuration for testing the README"""

import ops

import scenario


def pytest_markdown_docs_globals():
    class MyCharm(ops.CharmBase):
        META = {"name": "mycharm", "storage": {"foo": {"type": "filesystem"}}}

    class SimpleCharm(ops.CharmBase):
        META = {"name": "simplecharm"}

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, _: ops.StartEvent):
            pass

    class HistoryCharm(ops.CharmBase):
        META = {"name": "historycharm"}

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, _: ops.StartEvent):
            self.unit.set_workload_version("1")
            self.unit.set_workload_version("1.2")
            self.unit.set_workload_version("1.5")
            self.unit.set_workload_version("2.0")

    class PortCharm(ops.CharmBase):
        META = {"name": "portcharm"}

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)
            framework.observe(self.on.stop, self._on_stop)

        def _on_start(self, _: ops.StartEvent):
            self.unit.open_port(protocol="tcp", port=42)

        def _on_stop(self, _: ops.StopEvent):
            self.unit.close_port(protocol="tcp", port=42)

    return {
        "ops": ops,
        "scenario": scenario,
        "MyCharm": MyCharm,
        "HistoryCharm": HistoryCharm,
        "PortCharm": PortCharm,
        "SimpleCharm": SimpleCharm,
    }



import pytest

import ops
from ops.framework import StoredState as ops_storedstate

from scenario.state import State, StoredState
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(ops.CharmBase):
        META = {"name": "mycharm"}

        _read = {}
        _stored = ops_storedstate()
        _stored2 = ops_storedstate()

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self._stored.set_default(foo="bar", baz={12: 142})
            self._stored2.set_default(foo="bar", baz={12: 142})
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, _: ops.EventBase):
            self._read["foo"] = self._stored.foo
            self._read["baz"] = self._stored.baz

    return MyCharm


def test_stored_state_default(mycharm):
    out = trigger(State(), "start", mycharm, meta=mycharm.META)
    assert out.get_stored_state("_stored", owner_path="MyCharm").content == {
        "foo": "bar",
        "baz": {12: 142},
    }
    assert out.get_stored_state("_stored2", owner_path="MyCharm").content == {
        "foo": "bar",
        "baz": {12: 142},
    }


def test_stored_state_initialized(mycharm):
    out = trigger(
        State(
            stored_states={
                StoredState(
                    owner_path="MyCharm", name="_stored", content={"foo": "FOOX"}
                ),
            }
        ),
        "start",
        mycharm,
        meta=mycharm.META,
    )
    assert out.get_stored_state("_stored", owner_path="MyCharm").content == {
        "foo": "FOOX",
        "baz": {12: 142},
    }
    assert out.get_stored_state("_stored2", owner_path="MyCharm").content == {
        "foo": "bar",
        "baz": {12: 142},
    }


def test_positional_arguments():
    with pytest.raises(TypeError):
        StoredState("_stored", "")


def test_default_arguments():
    s = StoredState()
    assert s.name == "_stored"
    assert s.owner_path is None
    assert s.content == {}
    assert s._data_type_name == "StoredStateData"



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""
The :meth:`check_consistency` function is the primary entry point for the
consistency checks. Calling it ensures that the :class:`State` and the event,
in combination with the ``Context``, is viable in Juju. For example, Juju can't
emit a ``foo-relation-changed`` event on your charm unless your charm has
declared a ``foo`` relation endpoint in its metadata.

Normally, there is no need to explicitly call this function; that happens
automatically behind the scenes whenever you trigger an event.

If you have a clear false negative, are explicitly testing 'edge',
inconsistent situations, or for whatever reason the checker is in your way, you
can set the ``SCENARIO_SKIP_CONSISTENCY_CHECKS`` environment variable and skip
it altogether.
"""

import marshal
import os
import re
from collections import Counter, defaultdict
from collections.abc import Sequence
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from .errors import InconsistentScenarioError
from ._runtime import logger as scenario_logger
from .state import (
    CharmType,
    PeerRelation,
    SubordinateRelation,
    _Action,
    _CharmSpec,
    _normalise_name,
)

if TYPE_CHECKING:  # pragma: no cover
    from .state import State, _Event

logger = scenario_logger.getChild("consistency_checker")


class Results(NamedTuple):
    """Consistency checker return type.

    Each consistency check function returns a ``Results`` instance with the
    warnings and errors found during the check.
    """

    errors: Iterable[str]
    warnings: Iterable[str]


def check_consistency(
    state: "State",
    event: "_Event",
    charm_spec: "_CharmSpec[Any]",
    juju_version: str,
):
    """Validate the combination of a state, an event, a charm spec, and a Juju version.

    When invoked, it performs a series of checks that validate that the state is
    consistent with itself, with the event being emitted, the charm metadata,
    and so on.

    This function performs some basic validation of the combination of inputs
    that goes into a test and determines if the scenario is a
    realistic/plausible/consistent one.

    A scenario is inconsistent if it can practically never occur because it
    contradicts the Juju model. For example: Juju guarantees that upon calling
    ``config-get``, a charm will only ever get the keys it declared in its
    config metadata, so a :class:`scenario.State` declaring some config keys
    that are not in the charm's ``charmcraft.yaml`` is nonsense, and the
    combination of the two is inconsistent.
    """
    juju_version_: Tuple[int, ...] = tuple(map(int, juju_version.split(".")))

    if os.getenv("SCENARIO_SKIP_CONSISTENCY_CHECKS"):
        logger.info("skipping consistency checks.")
        return

    errors: List[str] = []
    warnings: List[str] = []

    checks: Tuple[Callable[..., Results]] = (
        check_containers_consistency,
        check_config_consistency,
        check_resource_consistency,
        check_event_consistency,
        check_secrets_consistency,
        check_storages_consistency,
        check_relation_consistency,
        check_network_consistency,
        check_cloudspec_consistency,
        check_storedstate_consistency,
    )  # type: ignore
    for check in checks:
        results = check(
            state=state,
            event=event,
            charm_spec=charm_spec,
            juju_version=juju_version_,
        )
        errors.extend(results.errors)
        warnings.extend(results.warnings)

    if errors:
        err_fmt = "\n".join(errors)
        raise InconsistentScenarioError(
            f"Inconsistent scenario. The following errors were found: {err_fmt}",
        )
    if warnings:
        err_fmt = "\n".join(warnings)
        logger.warning(
            f"This scenario is probably inconsistent. Double check, and ignore this "
            f"warning if you're sure. "
            f"The following warnings were found: {err_fmt}",
        )


def check_resource_consistency(
    *,
    state: "State",
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the internal consistency of the resources from metadata and in :class:`scenario.State`."""
    errors: List[str] = []
    warnings: List[str] = []

    resources_from_meta = set(charm_spec.meta.get("resources", {}))
    resources_from_state = {resource.name for resource in state.resources}
    if not resources_from_meta.issuperset(resources_from_state):
        errors.append(
            f"any and all resources passed to State.resources need to have been defined in "
            f"metadata.yaml. Metadata resources: {resources_from_meta}; "
            f"State.resources: {resources_from_state}.",
        )
    return Results(errors, warnings)


def check_event_consistency(
    *,
    event: "_Event",
    charm_spec: "_CharmSpec[CharmType]",
    state: "State",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the internal consistency of the ``_Event`` data structure.

    For example, it checks that a relation event has a relation instance, and that
    the relation endpoint name matches the event prefix.
    """
    errors: List[str] = []
    warnings: List[str] = []

    if not event._is_builtin_event(charm_spec):
        # This is a custom event - we can't make assumptions about its name and
        # semantics. It doesn't really make sense to do checks that are designed
        # for relations, workloads, and so on - most likely those will end up
        # with false positives. Realistically, we can't know about what the
        # requirements for the custom event are (in terms of the state), so we
        # skip everything here. Perhaps in the future, custom events could
        # optionally include some sort of state metadata that made testing
        # consistency possible?
        warnings.append(
            "this is a custom event; if its name makes it look like a builtin one "
            "(for example, a relation event, or a workload event), you might get some false-negative "
            "consistency checks.",
        )
        return Results(errors, warnings)

    if event._is_relation_event:
        _check_relation_event(charm_spec, event, state, errors, warnings)

    if event._is_workload_event:
        _check_workload_event(charm_spec, event, state, errors, warnings)

    if event._is_action_event:
        _check_action_event(charm_spec, event, state, errors, warnings)

    if event._is_storage_event:
        _check_storage_event(charm_spec, event, state, errors, warnings)

    return Results(errors, warnings)


def _check_relation_event(
    charm_spec: _CharmSpec[CharmType],  # noqa: U100
    event: "_Event",
    state: "State",
    errors: List[str],
    warnings: List[str],  # noqa: U100
):
    if not event.relation:
        errors.append(
            "cannot construct a relation event without the relation instance. "
            "Please pass one.",
        )
    else:
        if not event.name.startswith(_normalise_name(event.relation.endpoint)):
            errors.append(
                f"relation event should start with relation endpoint name. {event.name} does "
                f"not start with {event.relation.endpoint}.",
            )
        if event.relation not in state.relations:
            errors.append(
                f"cannot emit {event.name} because relation {event.relation.id} is not in the "
                f"state (a relation with the same ID is not sufficient - you must "
                f"pass the object in the state to the event).",
            )


def _check_workload_event(
    charm_spec: _CharmSpec[CharmType],  # noqa: U100
    event: "_Event",
    state: "State",
    errors: List[str],
    warnings: List[str],
):
    if not event.container:
        errors.append(
            "cannot construct a workload event without the container instance. "
            "Please pass one.",
        )
    else:
        if not event.name.startswith(_normalise_name(event.container.name)):
            errors.append(
                f"workload event should start with container name. {event.name} does "
                f"not start with {event.container.name}.",
            )
            if event.container not in state.containers:
                errors.append(
                    f"cannot emit {event.name} because container {event.container.name} "
                    f"is not in the state (a container with the same name is not "
                    f"sufficient - you must pass the object in the state to the event).",
                )
            if not event.container.can_connect:
                warnings.append(
                    "you **can** fire fire pebble-ready while the container cannot connect, "
                    "but that's most likely not what you want.",
                )
        names = Counter(exec.command_prefix for exec in event.container.execs)
        if dupes := [n for n in names if names[n] > 1]:
            errors.append(
                f"container {event.container.name} has duplicate command prefixes: {dupes}",
            )


def _check_action_event(
    charm_spec: _CharmSpec[CharmType],
    event: "_Event",
    state: "State",  # noqa: U100
    errors: List[str],
    warnings: List[str],
):
    action = event.action
    if not action:
        errors.append(
            "cannot construct a workload event without the container instance. "
            "Please pass one.",
        )
        return

    elif not event.name.startswith(_normalise_name(action.name)):
        errors.append(
            f"action event should start with action name. {event.name} does "
            f"not start with {action.name}.",
        )
    if action.name not in (charm_spec.actions or ()):
        errors.append(
            f"action event {event.name} refers to action {action.name} "
            f"which is not declared in the charm metadata (actions.yaml).",
        )
        return

    _check_action_param_types(charm_spec, action, errors, warnings)


def _check_storage_event(
    charm_spec: _CharmSpec[CharmType],
    event: "_Event",
    state: "State",
    errors: List[str],
    warnings: List[str],  # noqa: U100
):
    storage = event.storage
    meta = charm_spec.meta

    if not storage:
        errors.append(
            "cannot construct a storage event without the Storage instance. "
            "Please pass one.",
        )
    elif not event.name.startswith(_normalise_name(storage.name)):
        errors.append(
            f"storage event should start with storage name. {event.name} does "
            f"not start with {storage.name}.",
        )
    elif storage.name not in meta["storage"]:
        errors.append(
            f"storage event {event.name} refers to storage {storage.name} "
            f"which is not declared in the charm metadata (metadata.yaml) under 'storage'.",
        )
    elif storage not in state.storages:
        errors.append(
            f"cannot emit {event.name} because storage {storage.name} "
            f"is not in the state (an object with the same name and index is not "
            f"sufficient - you must pass the object in the state to the event).",
        )


def _check_action_param_types(
    charm_spec: _CharmSpec[CharmType],
    action: _Action,
    errors: List[str],
    warnings: List[str],
):
    actions = charm_spec.actions
    if not actions:
        return

    to_python_type = {
        "string": str,
        "boolean": bool,
        "integer": int,
        "number": Number,
        "array": Sequence,
        "object": dict,
    }
    expected_param_type: Dict[str, Any] = {}
    for par_name, par_spec in actions[action.name].get("params", {}).items():
        value = par_spec.get("type")
        if not value:
            errors.append(
                f"action parameter {par_name} has no type. "
                f"Charmcraft will be unhappy about this. ",
            )
            continue

        try:
            expected_param_type[par_name] = to_python_type[value]
        except KeyError:
            warnings.append(
                f"unknown data type declared for parameter {par_name}: type={value}. "
                f"Cannot consistency-check.",
            )

    for provided_param_name, provided_param_value in action.params.items():
        expected_type = expected_param_type.get(provided_param_name)
        if not expected_type:
            errors.append(
                f"param {provided_param_name} is not a valid parameter for {action.name}: "
                "missing from action specification",
            )
            continue
        if not isinstance(provided_param_value, expected_type):
            errors.append(
                f"param {provided_param_name} is of type {type(provided_param_value)}: "
                f"expecting {expected_type}",
            )


def check_storages_consistency(
    *,
    state: "State",
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of the :class:`scenario.State` storages with the charm_spec metadata."""
    state_storage = state.storages
    meta_storage = (charm_spec.meta or {}).get("storage", {})
    errors: List[str] = []

    if missing := {s.name for s in state_storage}.difference(
        set(meta_storage.keys()),
    ):
        errors.append(
            f"some storages passed to State were not defined in metadata.yaml: {missing}",
        )

    seen: List[Tuple[str, int]] = []
    for s in state_storage:
        tag = (s.name, s.index)
        if tag in seen:
            errors.append(
                f"duplicate storage in State: storage {s.name} with index {s.index} "
                f"occurs multiple times in State.storages.",
            )
        seen.append(tag)

    return Results(errors, [])


def _is_secret_identifier(value: Union[str, int, float, bool]) -> bool:
    """Return true iff the value is in the form `secret:{secret id}`."""
    # cf. https://github.com/juju/juju/blob/13eb9df3df16a84fd471af8a3c95ddbd04389b71/core/secrets/secret.go#L48
    return bool(re.match(r"secret:[0-9a-z]{20}$", str(value)))


def check_config_consistency(
    *,
    state: "State",
    charm_spec: "_CharmSpec[CharmType]",
    juju_version: Tuple[int, ...],
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of the :class:`scenario.State` config with the charm_spec config."""
    state_config = state.config
    meta_config = (charm_spec.config or {}).get("options", {})
    errors: List[str] = []

    for key, value in state_config.items():
        if key not in meta_config:
            errors.append(
                f"config option {key!r} in state.config but not specified in config.yaml or "
                f"charmcraft.yaml.",
            )
            continue

        converters = {
            "string": str,
            "int": int,
            "float": float,
            "boolean": bool,
        }
        if juju_version >= (3, 4):
            converters["secret"] = str

        validators = {
            "secret": _is_secret_identifier,
        }

        expected_type_name = meta_config[key].get("type", None)
        if not expected_type_name:
            errors.append(f"config.yaml invalid; option {key!r} has no 'type'.")
            continue
        validator = validators.get(expected_type_name)

        expected_type = converters.get(expected_type_name)
        if not expected_type:
            errors.append(
                f"config invalid for option {key!r}: 'type' {expected_type_name} unknown",
            )

        elif not isinstance(value, expected_type):
            errors.append(
                f"config invalid; option {key!r} should be of type {expected_type} "
                f"but is of type {type(value)}.",
            )

        elif validator and not validator(value):
            errors.append(
                f"config invalid: option {key!r} value {value!r} is not valid.",
            )

    return Results(errors, [])


def check_secrets_consistency(
    *,
    event: "_Event",
    state: "State",
    juju_version: Tuple[int, ...],
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of any :class:`scenario.Secret` in the :class:`scenario.State`."""
    errors: List[str] = []
    if not event._is_secret_event:
        return Results(errors, [])

    assert event.secret is not None
    if event.secret not in state.secrets:
        secret_key = event.secret.id if event.secret.id else event.secret.label
        errors.append(
            f"cannot emit {event.name} because secret {secret_key} is not in the state "
            f"(a secret with the same ID is not sufficient - you must pass the object "
            f"in the state to the event).",
        )
    elif juju_version < (3,):
        errors.append(
            f"secrets are not supported in the specified juju version {juju_version}. "
            f"Should be at least 3.0.",
        )

    return Results(errors, [])


def check_network_consistency(
    *,
    state: "State",
    event: "_Event",  # noqa: U100
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of any :class:`scenario.Network` in the :class:`scenario.State`."""
    errors: List[str] = []

    meta_bindings = set(charm_spec.meta.get("extra-bindings", ()))
    # add the implicit juju-info binding so we can override its network without
    # having to declare a relation for it in metadata
    implicit_bindings = {"juju-info"}
    all_relations = charm_spec.get_all_relations()
    non_sub_relations = {
        endpoint
        for endpoint, metadata in all_relations
        if metadata.get("scope") != "container"  # mark of a sub
    }

    state_bindings = {network.binding_name for network in state.networks}
    if diff := state_bindings.difference(
        meta_bindings.union(non_sub_relations).union(implicit_bindings),
    ):
        errors.append(
            f"Some network bindings defined in State are not in the metadata: {diff}.",
        )

    endpoints = {endpoint for endpoint, _ in all_relations}
    if collisions := endpoints.intersection(meta_bindings):
        errors.append(
            f"Extra bindings and relation endpoints cannot share the same name: {collisions}.",
        )

    return Results(errors, [])


def check_relation_consistency(
    *,
    state: "State",
    event: "_Event",  # noqa: U100
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of any relations in the :class:`scenario.State`."""
    errors: List[str] = []

    peer_relations_meta = charm_spec.meta.get("peers", {}).items()
    all_relations_meta = charm_spec.get_all_relations()

    def _get_relations(r: str):
        try:
            return state.get_relations(r)
        except ValueError:
            return ()

    # check relation types
    for endpoint, _ in peer_relations_meta:
        for relation in _get_relations(endpoint):
            if not isinstance(relation, PeerRelation):
                errors.append(
                    f"endpoint {endpoint} is a peer relation; "
                    f"expecting relation to be of type PeerRelation, got {type(relation)}",
                )

    known_endpoints = [a[0] for a in all_relations_meta]
    for relation in state.relations:
        if (ep := relation.endpoint) not in known_endpoints:
            errors.append(f"relation endpoint {ep} is not declared in metadata.")

    seen_ids: Set[int] = set()
    for endpoint, relation_meta in all_relations_meta:
        expected_sub = relation_meta.get("scope", "") == "container"
        relations = _get_relations(endpoint)
        for relation in relations:
            if relation.id in seen_ids:
                errors.append(
                    f"duplicate relation ID: {relation.id} is claimed "
                    f"by multiple Relation instances",
                )

            seen_ids.add(relation.id)
            is_sub = isinstance(relation, SubordinateRelation)
            if is_sub and not expected_sub:
                errors.append(
                    f"endpoint {endpoint} is not a subordinate relation; "
                    f"expecting relation to be of type Relation, "
                    f"got {type(relation)}",
                )
            if expected_sub and not is_sub:
                errors.append(
                    f"endpoint {endpoint} is not a subordinate relation; "
                    f"expecting relation to be of type SubordinateRelation, "
                    f"got {type(relation)}",
                )

    # check for duplicate endpoint names
    seen_endpoints: Set[str] = set()
    for endpoint, _ in all_relations_meta:
        if endpoint in seen_endpoints:
            errors.append("duplicate endpoint name in metadata.")
            break
        seen_endpoints.add(endpoint)

    return Results(errors, [])


def check_containers_consistency(
    *,
    state: "State",
    event: "_Event",
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the consistency of :class:`scenario.State` containers with the charm_spec metadata."""

    # event names will be normalized; need to compare against normalized container names.
    meta = charm_spec.meta
    meta_containers = list(map(_normalise_name, meta.get("containers", {})))
    state_containers = [_normalise_name(c.name) for c in state.containers]
    all_notices = {notice.id for c in state.containers for notice in c.notices}
    all_checks = {
        (c.name, check.name) for c in state.containers for check in c.check_infos
    }
    errors: List[str] = []

    # it's fine if you have containers in meta that are not in state.containers (yet), but it's
    # not fine if:
    # - you're processing a Pebble event and that container is not in state.containers or
    #   meta.containers
    if event._is_workload_event:
        evt_container_name = event.name.split("_pebble_")[0]
        if evt_container_name not in meta_containers:
            errors.append(
                f"the event being processed concerns container {evt_container_name!r}, but a "
                f"container with that name is not declared in the charm metadata",
            )
        if evt_container_name not in state_containers:
            errors.append(
                f"the event being processed concerns container {evt_container_name!r}, but a "
                f"container with that name is not present in the state. It's odd, but "
                f"consistent, if it cannot connect; but it should at least be there.",
            )
        # - you're processing a Notice event and that notice is not in any of the containers
        if event.notice and event.notice.id not in all_notices:
            errors.append(
                f"the event being processed concerns notice {event.notice!r}, but that "
                "notice is not in any of the containers present in the state.",
            )
        # - you're processing a Check event and that check is not in the check's container
        if (
            event.check_info
            and (evt_container_name, event.check_info.name) not in all_checks
        ):
            errors.append(
                f"the event being processed concerns check {event.check_info.name}, but that "
                f"check is not in the {evt_container_name} container.",
            )

    # - a container in state.containers is not in meta.containers
    if diff := (set(state_containers).difference(set(meta_containers))):
        errors.append(
            f"some containers declared in the state are not specified in metadata. "
            f"That's not possible. "
            f"Missing from metadata: {diff}.",
        )

    return Results(errors, [])


def check_cloudspec_consistency(
    *,
    state: "State",
    event: "_Event",
    charm_spec: "_CharmSpec[CharmType]",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check that Kubernetes models don't have :attr:`scenario.State.cloud_spec` set."""

    errors: List[str] = []
    warnings: List[str] = []

    if state.model.type == "kubernetes" and state.model.cloud_spec:
        errors.append(
            "CloudSpec is only available for machine charms, not Kubernetes charms. "
            "Simulate a machine substrate with: `State(..., model=Model(type='lxd'))`.",
        )

    return Results(errors, warnings)


def check_storedstate_consistency(
    *,
    state: "State",
    **_kwargs: Any,  # noqa: U101
) -> Results:
    """Check the internal consistency of any :class:`scenario.StoredState` in the :class:`scenario.State`."""
    errors: List[str] = []

    # Attribute names must be unique on each object.
    names: defaultdict[Optional[str], List[str]] = defaultdict(list)
    for ss in state.stored_states:
        names[ss.owner_path].append(ss.name)
    for owner, owner_names in names.items():
        if len(owner_names) != len(set(owner_names)):
            errors.append(
                f"{owner} has multiple StoredState objects with the same name.",
            )

    # The content must be marshallable.
    for ss in state.stored_states:
        # We don't need the marshalled state, just to know that it can be.
        # This is the same "only simple types" check that ops does.
        try:
            marshal.dumps(ss.content)
        except ValueError:
            errors.append(
                f"The StoredState object {ss.owner_path}.{ss.name} should contain only simple types.",
            )
    return Results(errors, [])



import datetime

import pytest
from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import ModelError
from ops.model import Secret as ops_Secret
from ops.model import SecretNotFoundError, SecretRotate

from scenario import Context
from scenario.state import Relation, Secret, State
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_get_secret_no_secret(mycharm):
    ctx = Context(mycharm, meta={"name": "local"})
    with ctx(ctx.on.update_status(), State()) as mgr:
        with pytest.raises(SecretNotFoundError):
            assert mgr.charm.model.get_secret(id="foo")
        with pytest.raises(SecretNotFoundError):
            assert mgr.charm.model.get_secret(label="foo")


@pytest.mark.parametrize("owner", ("app", "unit"))
def test_get_secret(mycharm, owner):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret({"a": "b"}, owner=owner)
    with ctx(
        state=State(secrets={secret}),
        event=ctx.on.update_status(),
    ) as mgr:
        assert mgr.charm.model.get_secret(id=secret.id).get_content()["a"] == "b"


@pytest.mark.parametrize("owner", ("app", "unit"))
def test_get_secret_get_refresh(mycharm, owner):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        tracked_content={"a": "b"},
        latest_content={"a": "c"},
        owner=owner,
    )
    with ctx(
        ctx.on.update_status(),
        State(secrets={secret}),
    ) as mgr:
        charm = mgr.charm
        assert (
            charm.model.get_secret(id=secret.id).get_content(refresh=True)["a"] == "c"
        )


@pytest.mark.parametrize("app", (True, False))
def test_get_secret_nonowner_peek_update(mycharm, app):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        tracked_content={"a": "b"},
        latest_content={"a": "c"},
    )
    with ctx(
        ctx.on.update_status(),
        State(
            leader=app,
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(id=secret.id).get_content()["a"] == "b"
        assert charm.model.get_secret(id=secret.id).peek_content()["a"] == "c"
        # Verify that the peek has not refreshed:
        assert charm.model.get_secret(id=secret.id).get_content()["a"] == "b"

        assert (
            charm.model.get_secret(id=secret.id).get_content(refresh=True)["a"] == "c"
        )
        assert charm.model.get_secret(id=secret.id).get_content()["a"] == "c"


@pytest.mark.parametrize("owner", ("app", "unit"))
def test_get_secret_owner_peek_update(mycharm, owner):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        tracked_content={"a": "b"},
        latest_content={"a": "c"},
        owner=owner,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(id=secret.id).get_content()["a"] == "b"
        assert charm.model.get_secret(id=secret.id).peek_content()["a"] == "c"
        # Verify that the peek has not refreshed:
        assert charm.model.get_secret(id=secret.id).get_content()["a"] == "b"
        assert (
            charm.model.get_secret(id=secret.id).get_content(refresh=True)["a"] == "c"
        )


@pytest.mark.parametrize("owner", ("app", "unit"))
def test_secret_changed_owner_evt_fails(mycharm, owner):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        tracked_content={"a": "b"},
        latest_content={"a": "c"},
        owner=owner,
    )
    with pytest.raises(ValueError):
        _ = ctx.on.secret_changed(secret)


@pytest.mark.parametrize(
    "evt_suffix,revision",
    [
        ("rotate", None),
        ("expired", 1),
        ("remove", 1),
    ],
)
def test_consumer_events_failures(mycharm, evt_suffix, revision):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        tracked_content={"a": "b"},
        latest_content={"a": "c"},
    )
    kwargs = {"secret": secret}
    if revision is not None:
        kwargs["revision"] = revision
    with pytest.raises(ValueError):
        _ = getattr(ctx.on, f"secret_{evt_suffix}")(**kwargs)


@pytest.mark.parametrize("app", (True, False))
def test_add(mycharm, app):
    ctx = Context(mycharm, meta={"name": "local"})
    with ctx(
        ctx.on.update_status(),
        State(leader=app),
    ) as mgr:
        charm = mgr.charm
        if app:
            charm.app.add_secret({"foo": "bar"}, label="mylabel")
        else:
            charm.unit.add_secret({"foo": "bar"}, label="mylabel")
        output = mgr.run()

    assert output.secrets
    secret = output.get_secret(label="mylabel")
    assert secret.latest_content == secret.tracked_content == {"foo": "bar"}
    assert secret.label == "mylabel"


def test_set_legacy_behaviour(mycharm):
    # in juju < 3.1.7, secret owners always used to track the latest revision.
    # ref: https://bugs.launchpad.net/juju/+bug/2037120
    ctx = Context(mycharm, meta={"name": "local"}, juju_version="3.1.6")
    rev1, rev2 = {"foo": "bar"}, {"foo": "baz", "qux": "roz"}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label="mylabel")
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev1
        )

        secret.set_content(rev2)
        # We need to get the secret again, because ops caches the content in
        # the object.
        secret: ops_Secret = charm.model.get_secret(label="mylabel")
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev2
        )

        state_out = mgr.run()

    assert (
        state_out.get_secret(label="mylabel").tracked_content
        == state_out.get_secret(label="mylabel").latest_content
        == rev2
    )


def test_set(mycharm):
    ctx = Context(mycharm, meta={"name": "local"})
    rev1, rev2 = {"foo": "bar"}, {"foo": "baz", "qux": "roz"}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label="mylabel")
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev1
        )

        # TODO: if this is done in the same event hook, it's more complicated
        # than this. Figure out what we should do here.
        # Also the next test, for Juju 3.3
        secret.set_content(rev2)
        assert secret.get_content() == rev1
        assert secret.peek_content() == secret.get_content(refresh=True) == rev2

        state_out = mgr.run()

    assert (
        state_out.get_secret(label="mylabel").tracked_content
        == state_out.get_secret(label="mylabel").latest_content
        == rev2
    )


def test_set_juju33(mycharm):
    ctx = Context(mycharm, meta={"name": "local"}, juju_version="3.3.1")
    rev1, rev2 = {"foo": "bar"}, {"foo": "baz", "qux": "roz"}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label="mylabel")
        assert secret.get_content() == rev1

        secret.set_content(rev2)
        assert secret.get_content() == rev1
        assert secret.peek_content() == rev2
        assert secret.get_content(refresh=True) == rev2

        state_out = mgr.run()

    assert (
        state_out.get_secret(label="mylabel").tracked_content
        == state_out.get_secret(label="mylabel").latest_content
        == rev2
    )


@pytest.mark.parametrize("app", (True, False))
def test_meta(mycharm, app):
    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        {"a": "b"},
        owner="app" if app else "unit",
        label="mylabel",
        description="foobarbaz",
        rotate=SecretRotate.HOURLY,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            leader=True,
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(label="mylabel")

        secret = charm.model.get_secret(id=secret.id)
        info = secret.get_info()

        assert secret.label is None
        assert info.label == "mylabel"
        assert info.rotation == SecretRotate.HOURLY


@pytest.mark.parametrize("leader", (True, False))
@pytest.mark.parametrize("owner", ("app", "unit", None))
def test_secret_permission_model(mycharm, leader, owner):
    expect_manage = bool(
        # if you're the leader and own this app secret
        (owner == "app" and leader)
        # you own this secret
        or (owner == "unit")
    )

    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        {"a": "b"},
        label="mylabel",
        owner=owner,
        description="foobarbaz",
        rotate=SecretRotate.HOURLY,
    )
    secret_id = secret.id
    with ctx(
        ctx.on.update_status(),
        State(
            leader=leader,
            secrets={secret},
        ),
    ) as mgr:
        # can always view
        secret: ops_Secret = mgr.charm.model.get_secret(id=secret_id)
        assert secret.get_content()["a"] == "b"
        assert secret.peek_content()
        assert secret.get_content(refresh=True)

        if expect_manage:
            assert secret.get_content()
            assert secret.peek_content()
            assert secret.get_content(refresh=True)

            assert secret.get_info()
            secret.set_content({"foo": "boo"})
            assert secret.get_content() == {"a": "b"}  # rev1!
            assert secret.get_content(refresh=True) == {"foo": "boo"}

            secret.remove_all_revisions()

        else:  # cannot manage
            # nothing else to do directly if you can't get a hold of the Secret instance
            # but we can try some raw backend calls
            with pytest.raises(ModelError):
                secret.get_info()

            with pytest.raises(ModelError):
                secret.set_content(content={"boo": "foo"})


@pytest.mark.parametrize("app", (True, False))
def test_grant(mycharm, app):
    ctx = Context(
        mycharm, meta={"name": "local", "requires": {"foo": {"interface": "bar"}}}
    )
    secret = Secret(
        {"a": "b"},
        owner="unit",
        label="mylabel",
        description="foobarbaz",
        rotate=SecretRotate.HOURLY,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            relations=[Relation("foo", "remote")],
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        secret = charm.model.get_secret(label="mylabel")
        foo = charm.model.get_relation("foo")
        if app:
            secret.grant(relation=foo)
        else:
            secret.grant(relation=foo, unit=foo.units.pop())
        output = mgr.run()
    vals = list(output.get_secret(label="mylabel").remote_grants.values())
    assert vals == [{"remote"}] if app else [{"remote/0"}]


def test_update_metadata(mycharm):
    exp = datetime.datetime(2050, 12, 12)

    ctx = Context(mycharm, meta={"name": "local"})
    secret = Secret(
        {"a": "b"},
        owner="unit",
        label="mylabel",
    )
    with ctx(
        ctx.on.update_status(),
        State(
            secrets={secret},
        ),
    ) as mgr:
        secret = mgr.charm.model.get_secret(label="mylabel")
        secret.set_info(
            label="babbuccia",
            description="blu",
            expire=exp,
            rotate=SecretRotate.DAILY,
        )
        output = mgr.run()

    secret_out = output.get_secret(label="babbuccia")
    assert secret_out.label == "babbuccia"
    assert secret_out.rotate == SecretRotate.DAILY
    assert secret_out.description == "blu"
    assert secret_out.expire == exp


@pytest.mark.parametrize("leader", (True, False))
def test_grant_after_add(leader):
    class GrantingCharm(CharmBase):
        def __init__(self, *args):
            super().__init__(*args)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, _):
            if leader:
                secret = self.app.add_secret({"foo": "bar"})
            else:
                secret = self.unit.add_secret({"foo": "bar"})
            secret.grant(self.model.relations["bar"][0])

    state = State(leader=leader, relations={Relation("bar")})
    ctx = Context(
        GrantingCharm, meta={"name": "foo", "provides": {"bar": {"interface": "bar"}}}
    )
    ctx.run(ctx.on.start(), state)


def test_grant_nonowner(mycharm):
    secret = Secret(
        {"a": "b"},
        label="mylabel",
        description="foobarbaz",
        rotate=SecretRotate.HOURLY,
    )
    secret_id = secret.id

    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(id=secret_id)
        secret = charm.model.get_secret(label="mylabel")
        foo = charm.model.get_relation("foo")

        with pytest.raises(ModelError):
            secret.grant(relation=foo)

    trigger(
        State(
            relations={Relation("foo", "remote")},
            secrets={secret},
        ),
        "update_status",
        mycharm,
        meta={"name": "local", "requires": {"foo": {"interface": "bar"}}},
        post_event=post_event,
    )


def test_add_grant_revoke_remove():
    class GrantingCharm(CharmBase):
        pass

    ctx = Context(
        GrantingCharm, meta={"name": "foo", "provides": {"bar": {"interface": "bar"}}}
    )
    relation_remote_app = "remote_secret_desirerer"
    relation_id = 42

    state = State(
        leader=True,
        relations={
            Relation("bar", remote_app_name=relation_remote_app, id=relation_id)
        },
    )

    with ctx(ctx.on.start(), state) as mgr:
        charm = mgr.charm
        secret = charm.app.add_secret({"foo": "bar"}, label="mylabel")
        bar_relation = charm.model.relations["bar"][0]

        secret.grant(bar_relation)
        output = mgr.run()

    assert output.secrets
    scenario_secret = output.get_secret(label="mylabel")
    assert relation_remote_app in scenario_secret.remote_grants[relation_id]

    with ctx(ctx.on.start(), output) as mgr:
        charm: GrantingCharm = mgr.charm
        secret = charm.model.get_secret(label="mylabel")
        secret.revoke(bar_relation)
        output = mgr.run()

    scenario_secret = output.get_secret(label="mylabel")
    assert scenario_secret.remote_grants == {}

    with ctx(ctx.on.start(), output) as mgr:
        charm: GrantingCharm = mgr.charm
        secret = charm.model.get_secret(label="mylabel")
        secret.remove_all_revisions()
        output = mgr.run()

    with pytest.raises(KeyError):
        output.get_secret(label="mylabel")


def test_secret_removed_event():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_remove, self._on_secret_remove)

        def _on_secret_remove(self, event):
            event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={"name": "foo"})
    secret = Secret({"a": "b"}, owner="app")
    old_revision = 42
    state = ctx.run(
        ctx.on.secret_remove(secret, revision=old_revision),
        State(leader=True, secrets={secret}),
    )
    assert secret in state.secrets
    assert ctx.removed_secret_revisions == [old_revision]


def test_secret_expired_event():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_expired, self._on_secret_expired)

        def _on_secret_expired(self, event):
            event.secret.set_content({"password": "newpass"})
            event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={"name": "foo"})
    secret = Secret({"password": "oldpass"}, owner="app")
    old_revision = 42
    state = ctx.run(
        ctx.on.secret_expired(secret, revision=old_revision),
        State(leader=True, secrets={secret}),
    )
    assert state.get_secret(id=secret.id).latest_content == {"password": "newpass"}
    assert ctx.removed_secret_revisions == [old_revision]


def test_remove_bad_revision():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_remove, self._on_secret_remove)

        def _on_secret_remove(self, event):
            with pytest.raises(ValueError):
                event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={"name": "foo"})
    secret = Secret({"a": "b"}, owner="app")
    ctx.run(
        ctx.on.secret_remove(secret, revision=secret._latest_revision),
        State(leader=True, secrets={secret}),
    )
    ctx.run(
        ctx.on.secret_remove(secret, revision=secret._tracked_revision),
        State(leader=True, secrets={secret}),
    )


def test_set_label_on_get():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, _):
            id = self.unit.add_secret({"foo": "bar"}).id
            secret = self.model.get_secret(id=id, label="label1")
            assert secret.label == "label1"
            secret = self.model.get_secret(id=id, label="label2")
            assert secret.label == "label2"

    ctx = Context(SecretCharm, meta={"name": "foo"})
    state = ctx.run(ctx.on.start(), State())
    assert state.get_secret(label="label2").tracked_content == {"foo": "bar"}


def test_no_additional_positional_arguments():
    with pytest.raises(TypeError):
        Secret({}, {}, None)


def test_default_values():
    contents = {"foo": "bar"}
    secret = Secret(contents)
    assert secret.latest_content == secret.tracked_content == contents
    assert secret.id.startswith("secret:")
    assert secret.label is None
    assert secret.description is None
    assert secret.owner is None
    assert secret.rotate is None
    assert secret.expire is None
    assert secret.remote_grants == {}



from unittest.mock import patch

import pytest
from ops import CharmBase

from scenario import Context, State
from scenario.state import _Event, _next_action_id


class MyCharm(CharmBase):
    pass


def test_run():
    ctx = Context(MyCharm, meta={"name": "foo"})
    state = State()

    with patch.object(ctx, "_run") as p:
        ctx._output_state = "foo"  # would normally be set within the _run call scope
        output = ctx.run(ctx.on.start(), state)
        assert output == "foo"

    assert p.called
    e = p.call_args.kwargs["event"]
    s = p.call_args.kwargs["state"]

    assert isinstance(e, _Event)
    assert e.name == "start"
    assert s is state


def test_run_action():
    ctx = Context(MyCharm, meta={"name": "foo"})
    state = State()
    expected_id = _next_action_id(update=False)

    with patch.object(ctx, "_run") as p:
        ctx._output_state = "foo"  # would normally be set within the _run call scope
        output = ctx.run(ctx.on.action("do-foo"), state)
        assert output == "foo"

    assert p.called
    e = p.call_args.kwargs["event"]
    s = p.call_args.kwargs["state"]

    assert isinstance(e, _Event)
    assert e.name == "do_foo_action"
    assert s is state
    assert e.action.id == expected_id


@pytest.mark.parametrize("app_name", ("foo", "bar", "george"))
@pytest.mark.parametrize("unit_id", (1, 2, 42))
def test_app_name(app_name, unit_id):
    ctx = Context(MyCharm, meta={"name": "foo"}, app_name=app_name, unit_id=unit_id)
    with ctx(ctx.on.start(), State()) as mgr:
        assert mgr.charm.app.name == app_name
        assert mgr.charm.unit.name == f"{app_name}/{unit_id}"


def test_context_manager():
    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"act": {}})
    state = State()
    with ctx(ctx.on.start(), state) as mgr:
        mgr.run()
        assert mgr.charm.meta.name == "foo"

    with ctx(ctx.on.action("act"), state) as mgr:
        mgr.run()
        assert mgr.charm.meta.name == "foo"



import pytest
from ops import CharmBase, Framework, StartEvent, StopEvent

from scenario import Context, State
from scenario.state import Port, StateValidationError, TCPPort, UDPPort


class MyCharm(CharmBase):
    META = {"name": "edgar"}

    def __init__(self, framework: Framework):
        super().__init__(framework)
        framework.observe(self.on.start, self._open_port)
        framework.observe(self.on.stop, self._close_port)

    def _open_port(self, _: StartEvent):
        self.unit.open_port("tcp", 12)

    def _close_port(self, _: StopEvent):
        assert self.unit.opened_ports()
        self.unit.close_port("tcp", 42)


@pytest.fixture
def ctx():
    return Context(MyCharm, meta=MyCharm.META)


def test_open_port(ctx):
    out = ctx.run(ctx.on.start(), State())
    assert len(out.opened_ports) == 1
    port = tuple(out.opened_ports)[0]

    assert port.protocol == "tcp"
    assert port.port == 12


def test_close_port(ctx):
    out = ctx.run(ctx.on.stop(), State(opened_ports={TCPPort(42)}))
    assert not out.opened_ports


def test_port_no_arguments():
    with pytest.raises(RuntimeError):
        Port()


@pytest.mark.parametrize("klass", (TCPPort, UDPPort))
def test_port_port(klass):
    with pytest.raises(StateValidationError):
        klass(port=0)
    with pytest.raises(StateValidationError):
        klass(port=65536)



from typing import Type

import pytest
from ops.charm import (
    CharmBase,
    CharmEvents,
    CollectStatusEvent,
    RelationBrokenEvent,
    RelationCreatedEvent,
    RelationDepartedEvent,
    RelationEvent,
)
from ops.framework import EventBase, Framework

from scenario import Context
from scenario.errors import UncaughtCharmError
from scenario.state import (
    _DEFAULT_JUJU_DATABAG,
    PeerRelation,
    Relation,
    RelationBase,
    State,
    StateValidationError,
    SubordinateRelation,
    _next_relation_id,
)
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharmEvents(CharmEvents):
        @classmethod
        def define_event(cls, event_kind: str, event_type: "Type[EventBase]"):
            if getattr(cls, event_kind, None):
                delattr(cls, event_kind)
            return super().define_event(event_kind, event_type)

    class MyCharm(CharmBase):
        _call = None
        called = False
        on = MyCharmEvents()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if self._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_get_relation(mycharm):
    def pre_event(charm: CharmBase):
        assert charm.model.get_relation("foo")
        assert charm.model.get_relation("bar") is None
        assert charm.model.get_relation("qux")
        assert charm.model.get_relation("zoo") is None

    trigger(
        State(
            config={"foo": "bar"},
            leader=True,
            relations={
                Relation(endpoint="foo", interface="foo", remote_app_name="remote"),
                Relation(endpoint="qux", interface="qux", remote_app_name="remote"),
            },
        ),
        "start",
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
                "bar": {"interface": "bar"},
            },
            "provides": {
                "qux": {"interface": "qux"},
                "zoo": {"interface": "zoo"},
            },
        },
        config={"options": {"foo": {"type": "string"}}},
        pre_event=pre_event,
    )


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
def test_relation_events(mycharm, evt_name, remote_app_name):
    relation = Relation(
        endpoint="foo", interface="foo", remote_app_name=remote_app_name
    )

    def callback(charm: CharmBase, e):
        if not isinstance(e, RelationEvent):
            return  # filter out collect status events

        if evt_name == "broken":
            assert charm.model.get_relation("foo") is None
            assert e.relation.app.name == remote_app_name
        else:
            assert charm.model.get_relation("foo").app.name == remote_app_name

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f"relation_{evt_name}",
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )


@pytest.mark.parametrize(
    "evt_name,has_unit",
    [
        ("changed", True),
        ("broken", False),
        ("departed", True),
        ("joined", True),
        ("created", False),
    ],
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
@pytest.mark.parametrize(
    "remote_unit_id",
    (0, 1),
)
def test_relation_events_attrs(
    mycharm, evt_name, has_unit, remote_app_name, remote_unit_id
):
    relation = Relation(
        endpoint="foo", interface="foo", remote_app_name=remote_app_name
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app
        if not isinstance(event, (RelationCreatedEvent, RelationBrokenEvent)):
            assert event.unit
        if isinstance(event, RelationDepartedEvent):
            assert event.departing_unit

    mycharm._call = callback

    ctx = Context(
        charm_type=mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )
    state = State(relations={relation})
    kwargs = {}
    if has_unit:
        kwargs["remote_unit"] = remote_unit_id
    event = getattr(ctx.on, f"relation_{evt_name}")(relation, **kwargs)
    ctx.run(event, state=state)


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
def test_relation_events_no_attrs(mycharm, evt_name, remote_app_name, caplog):
    relation = Relation(
        endpoint="foo",
        interface="foo",
        remote_app_name=remote_app_name,
        remote_units_data={0: {}, 1: {}},  # 2 units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        # .unit is always None for created and broken.
        if isinstance(event, (RelationCreatedEvent, RelationBrokenEvent)):
            assert event.unit is None
        else:
            assert event.unit
        assert (evt_name == "departed") is bool(getattr(event, "departing_unit", False))

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f"relation_{evt_name}",
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )

    if evt_name not in ("created", "broken"):
        assert (
            "remote unit ID unset, and multiple remote unit IDs are present"
            in caplog.text
        )


def test_relation_default_unit_data_regular():
    relation = Relation("baz")
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_units_data == {0: _DEFAULT_JUJU_DATABAG}


def test_relation_default_unit_data_sub():
    relation = SubordinateRelation("baz")
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_unit_data == _DEFAULT_JUJU_DATABAG


def test_relation_default_unit_data_peer():
    relation = PeerRelation("baz")
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
def test_relation_events_no_remote_units(mycharm, evt_name, caplog):
    relation = Relation(
        endpoint="foo",
        interface="foo",
        remote_units_data={},  # no units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        assert not event.unit

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f"relation_{evt_name}",
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )

    if evt_name not in ("created", "broken"):
        assert "remote unit ID unset; no remote unit data present" in caplog.text


@pytest.mark.parametrize("data", (set(), {}, [], (), 1, 1.0, None, b""))
def test_relation_unit_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        Relation(endpoint="foo", interface="foo", remote_units_data={0: {"a": data}})


@pytest.mark.parametrize("data", (set(), {}, [], (), 1, 1.0, None, b""))
def test_relation_app_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        Relation(endpoint="foo", interface="foo", local_app_data={"a": data})


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "relation",
    (Relation("a"), PeerRelation("b"), SubordinateRelation("c")),
)
def test_relation_event_trigger(relation, evt_name, mycharm):
    meta = {
        "name": "mycharm",
        "requires": {"a": {"interface": "i1"}},
        "provides": {
            "c": {
                "interface": "i3",
                # this is a subordinate relation.
                "scope": "container",
            }
        },
        "peers": {"b": {"interface": "i2"}},
    }
    trigger(
        State(relations={relation}),
        f"relation_{evt_name}",
        mycharm,
        meta=meta,
    )


def test_trigger_sub_relation(mycharm):
    meta = {
        "name": "mycharm",
        "provides": {
            "foo": {
                "interface": "bar",
                # this is a subordinate relation.
                "scope": "container",
            }
        },
    }

    sub1 = SubordinateRelation(
        "foo", remote_unit_data={"1": "2"}, remote_app_name="primary1"
    )
    sub2 = SubordinateRelation(
        "foo", remote_unit_data={"3": "4"}, remote_app_name="primary2"
    )

    def post_event(charm: CharmBase):
        b_relations = charm.model.relations["foo"]
        assert len(b_relations) == 2
        for relation in b_relations:
            assert len(relation.units) == 1

    trigger(
        State(relations={sub1, sub2}),
        "update_status",
        mycharm,
        meta=meta,
        post_event=post_event,
    )


def test_cannot_instantiate_relationbase():
    with pytest.raises(RuntimeError):
        RelationBase("")


def test_relation_ids():
    from scenario.state import _next_relation_id_counter

    initial_id = _next_relation_id_counter
    for i in range(10):
        rel = Relation("foo")
        assert rel.id == initial_id + i


def test_broken_relation_not_in_model_relations(mycharm):
    rel = Relation("foo")

    ctx = Context(
        mycharm, meta={"name": "local", "requires": {"foo": {"interface": "foo"}}}
    )
    with ctx(ctx.on.relation_broken(rel), state=State(relations={rel})) as mgr:
        charm = mgr.charm

        assert charm.model.get_relation("foo") is None
        assert charm.model.relations["foo"] == []


def test_get_relation_when_missing():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.update_status, self._on_update_status)
            self.framework.observe(self.on.config_changed, self._on_config_changed)
            self.relation = None

        def _on_update_status(self, _):
            self.relation = self.model.get_relation("foo")

        def _on_config_changed(self, _):
            self.relation = self.model.get_relation("foo", self.config["relation-id"])

    ctx = Context(
        MyCharm,
        meta={"name": "foo", "requires": {"foo": {"interface": "foo"}}},
        config={"options": {"relation-id": {"type": "int", "description": "foo"}}},
    )
    # There should be no error if the relation is missing - get_relation returns
    # None in that case.
    with ctx(ctx.on.update_status(), State()) as mgr:
        mgr.run()
        assert mgr.charm.relation is None

    # There should also be no error if the relation is present, of course.
    rel = Relation("foo")
    with ctx(ctx.on.update_status(), State(relations={rel})) as mgr:
        mgr.run()
        assert mgr.charm.relation.id == rel.id

    # If a relation that doesn't exist is requested, that should also not raise
    # an error.
    with ctx(ctx.on.config_changed(), State(config={"relation-id": 42})) as mgr:
        mgr.run()
        rel = mgr.charm.relation
        assert rel.id == 42
        assert not rel.active

    # If there's no defined relation with the name, then get_relation raises KeyError.
    ctx = Context(MyCharm, meta={"name": "foo"})
    with pytest.raises(UncaughtCharmError) as exc:
        ctx.run(ctx.on.update_status(), State())
    assert isinstance(exc.value.__cause__, KeyError)


@pytest.mark.parametrize("klass", (Relation, PeerRelation, SubordinateRelation))
def test_relation_positional_arguments(klass):
    with pytest.raises(TypeError):
        klass("foo", "bar", None)


def test_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = "database"
    interface = "postgresql"
    relation = Relation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_app_name == "remote"
    assert relation.limit == 1
    assert relation.remote_app_data == {}
    assert relation.remote_units_data == {0: _DEFAULT_JUJU_DATABAG}


def test_subordinate_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = "database"
    interface = "postgresql"
    relation = SubordinateRelation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_app_name == "remote"
    assert relation.remote_unit_id == 0
    assert relation.remote_app_data == {}
    assert relation.remote_unit_data == _DEFAULT_JUJU_DATABAG


def test_peer_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = "peers"
    interface = "shared"
    relation = PeerRelation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.peers_data == {0: _DEFAULT_JUJU_DATABAG}



import tempfile
from pathlib import Path

import pytest
import yaml
from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import ActiveStatus

from scenario import Context, State
from ..helpers import trigger


class MyCharm(CharmBase):
    META = {"name": "my-charm"}

    def __init__(self, framework: Framework):
        super().__init__(framework)
        foo = self.framework.charm_dir / "src" / "foo.bar"
        baz = self.framework.charm_dir / "src" / "baz" / "qux.kaboodle"

        self.unit.status = ActiveStatus(f"{foo.read_text()} {baz.read_text()}")


@pytest.fixture
def charm_virtual_root():
    with tempfile.TemporaryDirectory() as mycharm_virtual_root:
        t = Path(mycharm_virtual_root)
        src = t / "src"
        src.mkdir()
        foobar = src / "foo.bar"
        foobar.write_text("hello")

        baz = src / "baz"
        baz.mkdir(parents=True)
        quxcos = baz / "qux.kaboodle"
        quxcos.write_text("world")

        yield t


def test_charm_virtual_root(charm_virtual_root):
    out = trigger(
        State(),
        "start",
        charm_type=MyCharm,
        meta=MyCharm.META,
        charm_root=charm_virtual_root,
    )
    assert out.unit_status == ActiveStatus("hello world")


def test_charm_virtual_root_cleanup_if_exists(charm_virtual_root):
    meta_file = charm_virtual_root / "metadata.yaml"
    raw_ori_meta = yaml.safe_dump({"name": "karl"})
    meta_file.write_text(raw_ori_meta)

    ctx = Context(MyCharm, meta=MyCharm.META, charm_root=charm_virtual_root)
    with ctx(
        ctx.on.start(),
        State(),
    ) as mgr:
        assert meta_file.exists()
        assert meta_file.read_text() == yaml.safe_dump({"name": "my-charm"})
        assert (
            mgr.charm.meta.name == "my-charm"
        )  # not karl! Context.meta takes precedence
        mgr.run()
        assert meta_file.exists()

    # meta file was restored to its previous contents
    assert meta_file.read_text() == raw_ori_meta
    assert meta_file.exists()


def test_charm_virtual_root_cleanup_if_not_exists(charm_virtual_root):
    meta_file = charm_virtual_root / "metadata.yaml"

    assert not meta_file.exists()

    ctx = Context(MyCharm, meta=MyCharm.META, charm_root=charm_virtual_root)
    with ctx(
        ctx.on.start(),
        State(),
    ) as mgr:
        assert meta_file.exists()
        assert meta_file.read_text() == yaml.safe_dump({"name": "my-charm"})
        mgr.run()
        assert not meta_file.exists()

    assert not meta_file.exists()



#!/usr/bin/env python3
# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""Basic benchmarking charm.

Note that this is named benchmark_charm rather than charm as is typical to
avoid conflicts with ops.charm in the testing runs.
"""

import logging

import ops

logger = logging.getLogger("__name__")


class BenchmarkCharm(ops.CharmBase):
    """Charm the service."""

    _stored = ops.StoredState()

    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        framework.observe(self.on.update_status, self._on_update_status)
        framework.observe(self.on.stop, self._on_stop)
        framework.observe(self.on.config_changed, self._on_config_changed)

    def _on_update_status(self, _: ops.UpdateStatusEvent):
        # Say a bunch of things.
        for level in ("debug", "info", "warning", "error"):
            for i in range(50):
                getattr(logger, level)("This is message %s", i)

    def _on_stop(self, _: ops.StopEvent):
        """Do nothing - this exists to benchmark having an observer."""

    def _on_config_changed(self, event: ops.ConfigChangedEvent):
        event.defer()


if __name__ == "__main__":  # pragma: nocover
    ops.main(BenchmarkCharm)



import dataclasses
import datetime
import io
import tempfile
from pathlib import Path

import pytest
from ops import PebbleCustomNoticeEvent, pebble
from ops.charm import CharmBase
from ops.framework import Framework
from ops.pebble import ExecError, ServiceStartup, ServiceStatus

from scenario import Context
from scenario.state import CheckInfo, Container, Exec, Mount, Notice, State
from ..helpers import jsonpatch_delta, trigger


@pytest.fixture(scope="function")
def charm_cls():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_no_containers(charm_cls):
    def callback(self: CharmBase):
        assert not self.unit.containers

    trigger(
        State(),
        charm_type=charm_cls,
        meta={"name": "foo"},
        event="start",
        post_event=callback,
    )


def test_containers_from_meta(charm_cls):
    def callback(self: CharmBase):
        assert self.unit.containers
        assert self.unit.get_container("foo")

    trigger(
        State(),
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="start",
        post_event=callback,
    )


@pytest.mark.parametrize("can_connect", (True, False))
def test_connectivity(charm_cls, can_connect):
    def callback(self: CharmBase):
        assert can_connect == self.unit.get_container("foo").can_connect()

    trigger(
        State(containers={Container(name="foo", can_connect=can_connect)}),
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="start",
        post_event=callback,
    )


def test_fs_push(charm_cls):
    text = "lorem ipsum/n alles amat gloriae foo"
    file = tempfile.NamedTemporaryFile()
    pth = Path(file.name)
    pth.write_text(text)

    def callback(self: CharmBase):
        container = self.unit.get_container("foo")
        baz = container.pull("/bar/baz.txt")
        assert baz.read() == text

    trigger(
        State(
            containers={
                Container(
                    name="foo",
                    can_connect=True,
                    mounts={"bar": Mount(location="/bar/baz.txt", source=pth)},
                )
            }
        ),
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="start",
        post_event=callback,
    )


@pytest.mark.parametrize("make_dirs", (True, False))
def test_fs_pull(charm_cls, make_dirs):
    text = "lorem ipsum/n alles amat gloriae foo"

    def callback(self: CharmBase):
        container = self.unit.get_container("foo")
        if make_dirs:
            container.push("/foo/bar/baz.txt", text, make_dirs=make_dirs)
            # check that pulling immediately 'works'
            baz = container.pull("/foo/bar/baz.txt")
            assert baz.read() == text
        else:
            with pytest.raises(pebble.PathError):
                container.push("/foo/bar/baz.txt", text, make_dirs=make_dirs)

            # check that nothing was changed
            with pytest.raises((FileNotFoundError, pebble.PathError)):
                container.pull("/foo/bar/baz.txt")

    td = tempfile.TemporaryDirectory()
    container = Container(
        name="foo",
        can_connect=True,
        mounts={"foo": Mount(location="/foo", source=td.name)},
    )
    state = State(containers={container})

    ctx = Context(
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
    )
    with ctx(ctx.on.start(), state=state) as mgr:
        out = mgr.run()
        callback(mgr.charm)

    if make_dirs:
        # file = (out.get_container("foo").mounts["foo"].source + "bar/baz.txt").open("/foo/bar/baz.txt")

        # this is one way to retrieve the file
        file = Path(td.name + "/bar/baz.txt")

        # another is:
        assert (
            file
            == Path(out.get_container("foo").mounts["foo"].source) / "bar" / "baz.txt"
        )

        # but that is actually a symlink to the context's root tmp folder:
        assert (
            Path(ctx._tmp.name) / "containers" / "foo" / "foo" / "bar" / "baz.txt"
        ).read_text() == text
        assert file.read_text() == text

        # shortcut for API niceness purposes:
        file = container.get_filesystem(ctx) / "foo" / "bar" / "baz.txt"
        assert file.read_text() == text

    else:
        # nothing has changed
        out_purged = dataclasses.replace(out, stored_states=state.stored_states)
        assert not jsonpatch_delta(out_purged, state)


LS = """
.rw-rw-r--  228 ubuntu ubuntu 18 jan 12:05 -- charmcraft.yaml    
.rw-rw-r--  497 ubuntu ubuntu 18 jan 12:05 -- config.yaml        
.rw-rw-r--  900 ubuntu ubuntu 18 jan 12:05 -- CONTRIBUTING.md    
drwxrwxr-x    - ubuntu ubuntu 18 jan 12:06 -- lib                
.rw-rw-r--  11k ubuntu ubuntu 18 jan 12:05 -- LICENSE            
.rw-rw-r-- 1,6k ubuntu ubuntu 18 jan 12:05 -- metadata.yaml      
.rw-rw-r--  845 ubuntu ubuntu 18 jan 12:05 -- pyproject.toml     
.rw-rw-r--  831 ubuntu ubuntu 18 jan 12:05 -- README.md          
.rw-rw-r--   13 ubuntu ubuntu 18 jan 12:05 -- requirements.txt   
drwxrwxr-x    - ubuntu ubuntu 18 jan 12:05 -- src                
drwxrwxr-x    - ubuntu ubuntu 18 jan 12:05 -- tests              
.rw-rw-r-- 1,9k ubuntu ubuntu 18 jan 12:05 -- tox.ini            
"""
PS = """
    PID TTY          TIME CMD    
 298238 pts/3    00:00:04 zsh    
1992454 pts/3    00:00:00 ps     
"""


@pytest.mark.parametrize(
    "cmd, out",
    (
        ("ls", LS),
        ("ps", PS),
    ),
)
def test_exec(charm_cls, cmd, out):
    def callback(self: CharmBase):
        container = self.unit.get_container("foo")
        proc = container.exec([cmd])
        proc.wait()
        assert proc.stdout.read() == out

    trigger(
        State(
            containers={
                Container(
                    name="foo",
                    can_connect=True,
                    execs={Exec([cmd], stdout=out)},
                )
            }
        ),
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="start",
        post_event=callback,
    )


@pytest.mark.parametrize(
    "stdin,write",
    (
        [None, "hello world!"],
        ["hello world!", None],
        [io.StringIO("hello world!"), None],
    ),
)
def test_exec_history_stdin(stdin, write):
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.framework.observe(self.on.foo_pebble_ready, self._on_ready)

        def _on_ready(self, _):
            proc = self.unit.get_container("foo").exec(["ls"], stdin=stdin)
            if write:
                proc.stdin.write(write)
            proc.wait()

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    container = Container(name="foo", can_connect=True, execs={Exec([])})
    ctx.run(ctx.on.pebble_ready(container=container), State(containers={container}))
    assert ctx.exec_history[container.name][0].stdin == "hello world!"


def test_pebble_ready(charm_cls):
    def callback(self: CharmBase):
        foo = self.unit.get_container("foo")
        assert foo.can_connect()

    container = Container(name="foo", can_connect=True)

    trigger(
        State(containers={container}),
        charm_type=charm_cls,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="pebble_ready",
        post_event=callback,
    )


@pytest.mark.parametrize("starting_service_status", pebble.ServiceStatus)
def test_pebble_plan(charm_cls, starting_service_status):
    class PlanCharm(charm_cls):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_pebble_ready, self._on_ready)

        def _on_ready(self, event):
            foo = event.workload

            assert foo.get_plan().to_dict() == {
                "services": {"fooserv": {"startup": "enabled"}}
            }
            fooserv = foo.get_services("fooserv")["fooserv"]
            assert fooserv.startup == ServiceStartup.ENABLED
            assert fooserv.current == ServiceStatus.ACTIVE

            foo.add_layer(
                "bar",
                {
                    "summary": "bla",
                    "description": "deadbeef",
                    "services": {"barserv": {"startup": "disabled"}},
                },
            )

            foo.replan()
            assert foo.get_plan().to_dict() == {
                "services": {
                    "barserv": {"startup": "disabled"},
                    "fooserv": {"startup": "enabled"},
                }
            }

            assert foo.get_service("barserv").current == starting_service_status
            foo.start("barserv")
            # whatever the original state, starting a service sets it to active
            assert foo.get_service("barserv").current == ServiceStatus.ACTIVE

    container = Container(
        name="foo",
        can_connect=True,
        layers={
            "foo": pebble.Layer({
                "summary": "bla",
                "description": "deadbeef",
                "services": {"fooserv": {"startup": "enabled"}},
            })
        },
        service_statuses={
            "fooserv": pebble.ServiceStatus.ACTIVE,
            # todo: should we disallow setting status for services that aren't known YET?
            "barserv": starting_service_status,
        },
    )

    out = trigger(
        State(containers={container}),
        charm_type=PlanCharm,
        meta={"name": "foo", "containers": {"foo": {}}},
        event="pebble_ready",
    )

    def serv(name, obj):
        return pebble.Service(name, raw=obj)

    container = out.get_container(container.name)
    assert container.plan.services == {
        "barserv": serv("barserv", {"startup": "disabled"}),
        "fooserv": serv("fooserv", {"startup": "enabled"}),
    }
    assert container.services["fooserv"].current == pebble.ServiceStatus.ACTIVE
    assert container.services["fooserv"].startup == pebble.ServiceStartup.ENABLED

    assert container.services["barserv"].current == pebble.ServiceStatus.ACTIVE
    assert container.services["barserv"].startup == pebble.ServiceStartup.DISABLED


def test_exec_wait_error(charm_cls):
    state = State(
        containers={
            Container(
                name="foo",
                can_connect=True,
                execs={Exec(["foo"], stdout="hello pebble", return_code=1)},
            )
        }
    )

    ctx = Context(charm_cls, meta={"name": "foo", "containers": {"foo": {}}})
    with ctx(ctx.on.start(), state) as mgr:
        container = mgr.charm.unit.get_container("foo")
        proc = container.exec(["foo"])
        with pytest.raises(ExecError) as exc_info:
            proc.wait_output()
        assert exc_info.value.stdout == "hello pebble"


@pytest.mark.parametrize("command", (["foo"], ["foo", "bar"], ["foo", "bar", "baz"]))
def test_exec_wait_output(charm_cls, command):
    state = State(
        containers={
            Container(
                name="foo",
                can_connect=True,
                execs={Exec(["foo"], stdout="hello pebble", stderr="oepsie")},
            )
        }
    )

    ctx = Context(charm_cls, meta={"name": "foo", "containers": {"foo": {}}})
    with ctx(ctx.on.start(), state) as mgr:
        container = mgr.charm.unit.get_container("foo")
        proc = container.exec(command)
        out, err = proc.wait_output()
        assert out == "hello pebble"
        assert err == "oepsie"
        assert ctx.exec_history[container.name][0].command == command


def test_exec_wait_output_error(charm_cls):
    state = State(
        containers={
            Container(
                name="foo",
                can_connect=True,
                execs={Exec(["foo"], stdout="hello pebble", return_code=1)},
            )
        }
    )

    ctx = Context(charm_cls, meta={"name": "foo", "containers": {"foo": {}}})
    with ctx(ctx.on.start(), state) as mgr:
        container = mgr.charm.unit.get_container("foo")
        proc = container.exec(["foo"])
        with pytest.raises(ExecError):
            proc.wait_output()


def test_pebble_custom_notice(charm_cls):
    notices = [
        Notice(key="example.com/foo"),
        Notice(key="example.com/bar", last_data={"a": "b"}),
        Notice(key="example.com/baz", occurrences=42),
    ]
    container = Container(
        name="foo",
        can_connect=True,
        notices=notices,
    )

    state = State(containers=[container])
    ctx = Context(charm_cls, meta={"name": "foo", "containers": {"foo": {}}})
    with ctx(
        ctx.on.pebble_custom_notice(container=container, notice=notices[-1]), state
    ) as mgr:
        container = mgr.charm.unit.get_container("foo")
        assert container.get_notices() == [n._to_ops() for n in notices]


def test_pebble_custom_notice_in_charm():
    key = "example.com/test/charm"
    data = {"foo": "bar"}
    user_id = 100
    first_occurred = datetime.datetime(1979, 1, 25, 11, 0, 0)
    last_occured = datetime.datetime(2006, 8, 28, 13, 28, 0)
    last_repeated = datetime.datetime(2023, 9, 4, 9, 0, 0)
    occurrences = 42
    repeat_after = datetime.timedelta(days=7)
    expire_after = datetime.timedelta(days=365)

    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_pebble_custom_notice, self._on_custom_notice)

        def _on_custom_notice(self, event: PebbleCustomNoticeEvent):
            notice = event.notice
            assert notice.type == pebble.NoticeType.CUSTOM
            assert notice.key == key
            assert notice.last_data == data
            assert notice.user_id == user_id
            assert notice.first_occurred == first_occurred
            assert notice.last_occurred == last_occured
            assert notice.last_repeated == last_repeated
            assert notice.occurrences == occurrences
            assert notice.repeat_after == repeat_after
            assert notice.expire_after == expire_after

    notices = [
        Notice("example.com/test/other"),
        Notice("example.org/test/charm", last_data={"foo": "baz"}),
        Notice(
            key,
            last_data=data,
            user_id=user_id,
            first_occurred=first_occurred,
            last_occurred=last_occured,
            last_repeated=last_repeated,
            occurrences=occurrences,
            repeat_after=repeat_after,
            expire_after=expire_after,
        ),
    ]
    container = Container(
        name="foo",
        can_connect=True,
        notices=notices,
    )
    state = State(containers=[container])
    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    ctx.run(ctx.on.pebble_custom_notice(container=container, notice=notices[-1]), state)


def test_pebble_check_failed():
    infos = []

    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_pebble_check_failed, self._on_check_failed)

        def _on_check_failed(self, event):
            infos.append(event.info)

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    check = CheckInfo("http-check", failures=7, status=pebble.CheckStatus.DOWN)
    container = Container("foo", check_infos={check})
    state = State(containers={container})
    ctx.run(ctx.on.pebble_check_failed(container, check), state=state)
    assert len(infos) == 1
    assert infos[0].name == "http-check"
    assert infos[0].status == pebble.CheckStatus.DOWN
    assert infos[0].failures == 7


def test_pebble_check_recovered():
    infos = []

    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(
                self.on.foo_pebble_check_recovered, self._on_check_recovered
            )

        def _on_check_recovered(self, event):
            infos.append(event.info)

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    check = CheckInfo("http-check")
    container = Container("foo", check_infos={check})
    state = State(containers={container})
    ctx.run(ctx.on.pebble_check_recovered(container, check), state=state)
    assert len(infos) == 1
    assert infos[0].name == "http-check"
    assert infos[0].status == pebble.CheckStatus.UP
    assert infos[0].failures == 0


def test_pebble_check_failed_two_containers():
    foo_infos = []
    bar_infos = []

    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(
                self.on.foo_pebble_check_failed, self._on_foo_check_failed
            )
            framework.observe(
                self.on.bar_pebble_check_failed, self._on_bar_check_failed
            )

        def _on_foo_check_failed(self, event):
            foo_infos.append(event.info)

        def _on_bar_check_failed(self, event):
            bar_infos.append(event.info)

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}, "bar": {}}})
    check = CheckInfo("http-check", failures=7, status=pebble.CheckStatus.DOWN)
    foo_container = Container("foo", check_infos={check})
    bar_container = Container("bar", check_infos={check})
    state = State(containers={foo_container, bar_container})
    ctx.run(ctx.on.pebble_check_failed(foo_container, check), state=state)
    assert len(foo_infos) == 1
    assert foo_infos[0].name == "http-check"
    assert foo_infos[0].status == pebble.CheckStatus.DOWN
    assert foo_infos[0].failures == 7
    assert len(bar_infos) == 0


def test_pebble_add_layer():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.foo_pebble_ready, self._on_foo_ready)

        def _on_foo_ready(self, _):
            self.unit.get_container("foo").add_layer(
                "foo",
                {"checks": {"chk1": {"override": "replace"}}},
            )

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    container = Container("foo", can_connect=True)
    state_out = ctx.run(
        ctx.on.pebble_ready(container), state=State(containers={container})
    )
    chk1_info = state_out.get_container("foo").get_check_info("chk1")
    assert chk1_info.status == pebble.CheckStatus.UP


def test_pebble_start_check():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.foo_pebble_ready, self._on_foo_ready)
            framework.observe(self.on.config_changed, self._on_config_changed)

        def _on_foo_ready(self, _):
            container = self.unit.get_container("foo")
            container.add_layer(
                "foo",
                {"checks": {"chk1": {"override": "replace", "startup": "disabled"}}},
            )

        def _on_config_changed(self, _):
            container = self.unit.get_container("foo")
            container.start_checks("chk1")

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    container = Container("foo", can_connect=True)

    # Ensure that it starts as inactive.
    state_out = ctx.run(
        ctx.on.pebble_ready(container), state=State(containers={container})
    )
    chk1_info = state_out.get_container("foo").get_check_info("chk1")
    assert chk1_info.status == pebble.CheckStatus.INACTIVE

    # Verify that start_checks works.
    state_out = ctx.run(ctx.on.config_changed(), state=state_out)
    chk1_info = state_out.get_container("foo").get_check_info("chk1")
    assert chk1_info.status == pebble.CheckStatus.UP


def test_pebble_stop_check():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.config_changed, self._on_config_changed)

        def _on_config_changed(self, _):
            container = self.unit.get_container("foo")
            container.stop_checks("chk1")

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    info_in = CheckInfo("chk1", status=pebble.CheckStatus.UP)
    container = Container("foo", can_connect=True, check_infos=frozenset({info_in}))
    state_out = ctx.run(ctx.on.config_changed(), state=State(containers={container}))
    info_out = state_out.get_container("foo").get_check_info("chk1")
    assert info_out.status == pebble.CheckStatus.INACTIVE


def test_pebble_replan_checks():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.config_changed, self._on_config_changed)

        def _on_config_changed(self, _):
            container = self.unit.get_container("foo")
            container.replan()

    ctx = Context(MyCharm, meta={"name": "foo", "containers": {"foo": {}}})
    info_in = CheckInfo("chk1", status=pebble.CheckStatus.INACTIVE)
    layer = pebble.Layer({"checks": {"chk1": {"override": "replace"}}})
    container = Container(
        "foo",
        can_connect=True,
        check_infos=frozenset({info_in}),
        layers={"layer1": layer},
    )
    state_out = ctx.run(ctx.on.config_changed(), state=State(containers={container}))
    info_out = state_out.get_container("foo").get_check_info("chk1")
    assert info_out.status == pebble.CheckStatus.UP



#!/usr/bin/env python3
# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

import pathlib

import ops
import pytest

from scenario import Context, Resource, State


class ResourceCharm(ops.CharmBase):
    def __init__(self, framework):
        super().__init__(framework)


def test_get_resource():
    ctx = Context(
        ResourceCharm,
        meta={
            "name": "resource-charm",
            "resources": {"foo": {"type": "file"}, "bar": {"type": "file"}},
        },
    )
    resource1 = Resource(name="foo", path=pathlib.Path("/tmp/foo"))
    resource2 = Resource(name="bar", path=pathlib.Path("~/bar"))
    with ctx(
        ctx.on.update_status(), state=State(resources={resource1, resource2})
    ) as mgr:
        assert mgr.charm.model.resources.fetch("foo") == resource1.path
        assert mgr.charm.model.resources.fetch("bar") == resource2.path
        with pytest.raises(NameError):
            mgr.charm.model.resources.fetch("baz")
